Require Export real_facts.

Module COMPARE_CORRECT(Nums : NUMS).

Module Real_facts := REAL_FACTS(Nums).
Export Real_facts.

Theorem Scompare2_le : 
  forall n x y vx vy, 
    Srepresents x vx ->
    Srepresents y vy ->
    Scompare2 x y n = Lt ->
    vx <= vy.
Proof.
assert (Hbpos :=  base_positive).
intros n; induction n as [| n IHn]; simpl.
inversion 3.
intros x y vx vy Hrepx Hrepy.
destruct x as [x1 x']; destruct y as [y1 y'];
  inversion Hrepx; inversion Hrepy; subst.

compare_tac G1.

intros Hcomp.
replace (IDR x1) with (IDR y1) by IDR_to_Z.
apply Rdiv_le_compat_r; [IDR_to_Z|].
apply Rplus_le_compat_l.
generalize Hcomp; apply IHn; assumption.

compare_eq_tac G2; [| intros _ ].

destruct x' as [x2 x''];
  inversion_to_variable x''; subst.

compare_lt_tac G3; [ intros _ | ].

apply Rdiv_le_compat_r; [IDR_to_Z|].
replace (IDR y1) with (IDR x1 + 1) by IDR_to_Z.
rewrite (Rplus_assoc (IDR x1)); apply Rplus_le_compat_l.
apply Rle_trans with 0; [|dec_tac; fourier].
apply Rle_div_r; [IDR_to_Z|].
assert (U1 : IDR x2 <= -1) by IDR_to_Z.
dec_tac; fourier.

destruct y' as [y2 y''];
  inversion_to_variable y''; subst.
compare_gt_tac G4;
intros Hcomp;
  (apply Rdiv_le_compat_r; [IDR_to_Z|]);
  replace (IDR y1) with (IDR x1 + 1) by IDR_to_Z;
    rewrite (Rplus_assoc (IDR x1));
      apply Rplus_le_compat_l.

replace ( 1 + (IDR y2 + r) / Z2R b) with 
  ((IDR y2 + Z2R b+ r) / Z2R b) by field_IDR.
apply Rdiv_le_compat_r; [IDR_to_Z|].
assert (IDR x2 <= IDR y2 + Z2R b - 2) by IDR_to_Z; 
  dec_tac; fourier.

replace (1 + (IDR y2 + r) / Z2R b) with 
  (1/Z2R b + (IDR (y2+Dmax)%D + r)/ Z2R b); [ | field_IDR].
replace ((IDR x2 + r1) / Z2R b) with 
  (1/Z2R b + (IDR (x2-1)%D + r1)/ Z2R b); [ | field_IDR].
apply Rplus_le_compat_l.
generalize Hcomp; apply IHn.
constructor; assumption || simpl_IDZ; omega.
constructor; assumption || simpl_IDZ; omega.

apply Rdiv_le_compat_r; [IDR_to_Z|].
assert (IDR x1 + 2 <= IDR y1) by IDR_to_Z; 
  dec_tac; fourier.

compare_eq_tac G2; [| inversion 1].

destruct x' as [x2 x'']; 
  inversion_to_variable x''; subst.
compare_gt_tac G3; [inversion 1|].

destruct y' as [y2 y''];
  inversion_to_variable y''; subst.
compare_lt_tac G4; [inversion 1|].

intros Hcomp.
apply Rdiv_le_compat_r; [IDR_to_Z|].
replace (IDR x1) with (IDR y1 + 1) by IDR_to_Z.
rewrite (Rplus_assoc (IDR y1)).
apply Rplus_le_compat_l.
replace (1 + (IDR x2 + r1) / Z2R b) with 
  (1/Z2R b + (IDR (x2+Dmax)%D + r1)/ Z2R b); [ | field_IDR].
replace ((IDR y2 + r) / Z2R b) with 
  (1/Z2R b + (IDR (y2-1)%D + r)/ Z2R b); [ | field_IDR].
apply Rplus_le_compat_l.
generalize Hcomp; apply IHn.
constructor; assumption || simpl_IDZ; omega.
constructor; assumption || simpl_IDZ; omega.
Qed.

Theorem Scompare2_gt_lt : 
  forall n x y, 
  Scompare2 x y n = Gt ->
  Scompare2 y x n = Lt.
Proof.
intros n; induction n; simpl.
inversion 3.
intros x y.
destruct x as [x1 x']; destruct y as [y1 y'].
destruct x' as [x2 x'']; destruct y' as [y2 y''].

do 6 
(let H := fresh "G" in 
compare_tac H; 
match goal with 
|- Lt = Gt -> _ => inversion 1
| _ => apply IHn ||
           (absurd (IDZ x1 = IDZ y1); omega) || 
           (absurd (IDZ x2 = 0)%Z; omega) || 
           (absurd (IDZ y2 = 0)%Z; omega) || 
           auto
end).
Qed.

Theorem Scompare2_ge : 
  forall n x y vx vy, 
    Srepresents x vx ->
    Srepresents y vy ->
    Scompare2 x y n = Gt ->
    vx >= vy.
Proof.
intros n x y vx vy Hrepx Hrepy Hcomp.
apply Rle_ge.
cut (Scompare2 y x n = Lt).
apply Scompare2_le; assumption.
apply Scompare2_gt_lt; assumption.
Qed.

Theorem Scompare2_eq : 
  forall n x y vx vy, 
    Srepresents x vx ->
    Srepresents y vy ->
    Scompare2 x y n = Eq ->
    -2 <= (vx - vy) * Z2R b^ n <= 2.
Proof.
assert (Hbpos :=  base_positive).
intros n; induction n as [| n IHn]; simpl;
  intros x y vx vy Hrepx Hrepy.

intros _.
rewrite Rmult_1_r.
assert (U1 := Srepresents_interval _ _ Hrepx ).
assert (U2 := Srepresents_interval _ _ Hrepy ).
dec_tac; split; fourier.

destruct x as [x1 x']; destruct y as [y1 y'];
  inversion Hrepx; inversion Hrepy; subst.

replace (((IDR x1 + r) / Z2R b - (IDR y1 + r0) / Z2R b) * (Z2R b * Z2R b ^ n)) with 
   (((IDR x1 + r)  - (IDR y1 + r0) ) * Z2R b ^ n) by field_IDR.
compare_tac G1.

replace (IDR x1) with (IDR y1) by IDR_to_Z.
replace (IDR y1 + r - (IDR y1 + r0)) with 
  (r - r0) by ring.
apply IHn; assumption.

compare_eq_tac G2;[|inversion 1].

replace (IDR y1) with (IDR x1 + 1) by IDR_to_Z.
replace (IDR x1 + r - (IDR x1 + 1 + r0)) with 
  (r - (1 + r0)) by ring.
destruct x' as [x2 x''];
  inversion_to_variable x''; subst.
compare_lt_tac G3; [inversion 1|].
destruct y' as [y2 y''];
  inversion_to_variable y''; subst.
compare_gt_tac G4; [inversion 1|].
replace (((IDR x2 + r1) / Z2R b - (1 + (IDR y2 + r) / Z2R b)) * Z2R b ^ n) with 
  (((IDR (x2-1)%D + r1) / Z2R b - ((IDR (y2 +Dmax)%D+ r) / Z2R b)) * Z2R b ^ n) by field_IDR.
apply IHn;
  constructor; assumption || simpl_IDZ; omega.

compare_eq_tac G2;[| inversion 1].

replace (IDR x1) with (IDR y1 + 1) by IDR_to_Z.
replace (IDR y1 + 1 + r - (IDR y1 + r0)) with 
  (r +1 - r0) by ring.
destruct x' as [x2 x''];
  inversion_to_variable x''; subst.
compare_gt_tac G3; [inversion 1|].
destruct y' as [y2 y''];
  inversion_to_variable y''; subst.
compare_lt_tac G4; [inversion 1|].
replace (((IDR x2 + r1) / Z2R b + 1- (IDR y2 + r) / Z2R b) * Z2R b ^ n) with 
  (((IDR (x2+Dmax)%D + r1) / Z2R b - ((IDR (y2 -1)%D+ r) / Z2R b)) * Z2R b ^ n) by field_IDR.
apply IHn;
  constructor; assumption || simpl_IDZ; omega.
Qed.

Theorem Scompare_lt : 
  forall n x y vx vy, 
    Srepresents x vx ->
    Srepresents y vy ->
    Scompare x y n = Lt ->
    vx < vy.
Proof.
assert (Hbpos :=  base_positive).
intros n; induction n as [| n IHn]; simpl.
inversion 3.
intros x y vx vy Hrepx Hrepy.
destruct x as [x1 x']; destruct y as [y1 y'];
  inversion Hrepx; inversion Hrepy; subst.

compare_tac G1.

intros Hcomp.
replace (IDR x1) with (IDR y1) by IDR_to_Z.
apply Rdiv_lt_compat_r; [IDR_to_Z|].
apply Rplus_lt_compat_l.
generalize Hcomp; apply IHn; assumption.

compare_tac G2; [| intros _ |].

destruct x' as [x2 x''];
  inversion_to_variable x''; subst.

compare_eq_tac G3; [| intros _ ].

destruct y' as [y2 y''];
  inversion_to_variable y''; subst.
compare_eq_tac G4;
intros Hcomp;
  (apply Rdiv_lt_compat_r; [IDR_to_Z|]);
  replace (IDR y1) with (IDR x1 + 2) by IDR_to_Z;
    rewrite (Rplus_assoc (IDR x1));
      apply Rplus_lt_compat_l;
	replace (IDR x2) with (Z2R b - 1) by IDR_to_Z.

replace ((Z2R b - 1 + r1) / Z2R b) with 
  (1 + (IDR (-(1))%D + r1)/ Z2R b); [ | field_IDR].
replace (IDR y2) with (- Z2R b + 1) by IDR_to_Z.
replace (2 + (- Z2R b + 1 + r) / Z2R b) with 
  (1 + (IDR 1%D + r)/ Z2R b); [ | field_IDR].
apply Rplus_lt_compat_l.
generalize Hcomp; apply IHn.
constructor; assumption || simpl_IDZ; omega.
constructor; assumption || simpl_IDZ; omega.

replace ( 2 + (IDR y2 + r) / Z2R b) with 
  ((2* Z2R b + IDR y2 + r) / Z2R b) by field_IDR.
apply Rdiv_lt_compat_r; [IDR_to_Z|].
assert (U1 : - Z2R b + 2 <= IDR y2) by IDR_to_Z.
dec_tac; fourier.

apply Rdiv_lt_compat_r; [IDR_to_Z|].
replace (IDR y1) with (IDR x1 + 2) by IDR_to_Z.
rewrite (Rplus_assoc (IDR x1)); apply Rplus_lt_compat_l.
apply Rlt_le_trans with 1; [|dec_tac; fourier].
apply Rlt_div_r; [IDR_to_Z|].
assert (U1 : IDR x2 <= Z2R b - 2) by IDR_to_Z.
dec_tac; fourier.

apply Rdiv_lt_compat_r; [IDR_to_Z|].
assert (IDR x1 + 3 <= IDR y1) by IDR_to_Z; 
  dec_tac; fourier.

destruct x' as [x2 x''];
  inversion_to_variable x''; subst.
compare_lt_tac G3;[intros _|].

apply Rdiv_lt_compat_r; [IDR_to_Z|].
replace (IDR y1) with (IDR x1 + 1) by IDR_to_Z.
rewrite (Rplus_assoc (IDR x1)); apply Rplus_lt_compat_l.
apply Rlt_le_trans with 0; [|dec_tac; fourier].
apply Rlt_div_r; [IDR_to_Z|].
assert (U1 : IDR x2 <= -2) by IDR_to_Z.
dec_tac; fourier.

destruct y' as [y2 y''];
  inversion_to_variable y''; subst.
compare_gt_tac G4;
intros Hcomp;
  (apply Rdiv_lt_compat_r; [IDR_to_Z|]);
  replace (IDR y1) with (IDR x1 + 1) by IDR_to_Z;
    rewrite (Rplus_assoc (IDR x1));
      apply Rplus_lt_compat_l.

replace ( 1 + (IDR y2 + r) / Z2R b) with 
  ((IDR y2 + Z2R b+ r) / Z2R b) by field_IDR.
apply Rdiv_lt_compat_r; [IDR_to_Z|].
assert (IDR x2 <= IDR y2 + Z2R b - 3) by IDR_to_Z; 
  dec_tac; fourier.

replace (1 + (IDR y2 + r) / Z2R b) with 
  (2/Z2R b + (IDR (y2+Dmax -1)%D + r)/ Z2R b); [ | field_IDR].
replace ((IDR x2 + r1) / Z2R b) with 
  (2/Z2R b + (IDR (x2-2)%D + r1)/ Z2R b); [ | field_IDR].
apply Rplus_lt_compat_l.
generalize Hcomp; apply IHn.
constructor; assumption || simpl_IDZ; omega.
constructor; assumption || simpl_IDZ; omega.

compare_tac G2; [| | inversion 1].

destruct x' as [x2 x'']; 
  inversion_to_variable x''; subst.
compare_eq_tac G3; [| inversion 1].

destruct y' as [y2 y''];
  inversion_to_variable y''; subst.
compare_eq_tac G4; [| inversion 1].

intros Hcomp.
apply Rdiv_lt_compat_r; [IDR_to_Z|].
replace (IDR x1) with (IDR y1 + 2) by IDR_to_Z.
rewrite (Rplus_assoc (IDR y1)).
apply Rplus_lt_compat_l.
replace (IDR x2) with (- Z2R b + 1) by IDR_to_Z.
replace (2 + (- Z2R b + 1 + r1) / Z2R b) with 
  (1 + (IDR 1%D + r1)/ Z2R b); [ | field_IDR].
replace (IDR y2) with (Z2R b -1) by IDR_to_Z.
replace ((Z2R b - 1 + r) / Z2R b) with 
  (1 + (IDR (-(1))%D + r)/ Z2R b); [ | field_IDR].
apply Rplus_lt_compat_l.
generalize Hcomp; apply IHn.
constructor; assumption || simpl_IDZ; omega.
constructor; assumption || simpl_IDZ; omega.

destruct x' as [x2 x''];
  inversion_to_variable x''; subst.

compare_gt_tac G3; [ inversion 1|].

destruct y' as [y2 y''];
  inversion_to_variable y''; subst.
compare_lt_tac G4; [ inversion 1|].

intros Hcomp.
apply Rdiv_lt_compat_r; [IDR_to_Z|].
replace (IDR x1) with (IDR y1 + 1) by IDR_to_Z.
rewrite (Rplus_assoc (IDR y1)).
apply Rplus_lt_compat_l.
replace (1 + (IDR x2 + r1) / Z2R b) with 
  (2/Z2R b + (IDR (x2+Dmax-1)%D + r1)/ Z2R b); [ | field_IDR].
replace ((IDR y2 + r) / Z2R b) with 
  (2/Z2R b + (IDR (y2-2)%D + r)/ Z2R b); [ | field_IDR].
apply Rplus_lt_compat_l.
generalize Hcomp; apply IHn.
constructor; assumption || simpl_IDZ; omega.
constructor; assumption || simpl_IDZ; omega.
Qed.

Theorem Scompare_gt_lt : 
  forall n x y,
    Scompare x y n = Gt ->
    Scompare y x n = Lt.
Proof.
intros n; induction n as [| n IHn]; simpl.
inversion 3.
intros x y.
destruct x as [x1 x']; destruct y as [y1 y'].
destruct x' as [x2 x'']; destruct y' as [y2 y''].

do 6 
(let H := fresh "G" in 
compare_tac H; 
match goal with 
|- Lt = Gt -> _ => inversion 1
| _ => apply IHn ||
           (absurd (IDZ x1 = IDZ y1); omega) || 
           (absurd (IDZ x2 = 0)%Z; omega) || 
           (absurd (IDZ x2 = b-1)%Z; omega) || 
           (absurd (IDZ y2 = 0)%Z; omega) || 
           (absurd (IDZ y2 = b-1)%Z; omega) || 
           auto
end).
Qed.

Theorem Scompare_gt : 
  forall n x y vx vy, 
    Srepresents x vx ->
    Srepresents y vy ->
    Scompare x y n = Gt ->
    vx > vy.
Proof.
intros n x y vx vy Hrepx Hrepy Hcomp.
apply Rlt_gt.
cut (Scompare y x n = Lt).
apply Scompare_lt; assumption.
apply Scompare_gt_lt; assumption.
Qed.

Theorem Scompare_eq : 
  forall n x y vx vy, 
    Srepresents x vx ->
    Srepresents y vy ->
    Scompare x y n = Eq ->
    -2 <= (vx - vy) * Z2R b^ n<= 2.
Proof.
assert (Hbpos :=  base_positive).
intros n; induction n as [| n IHn]; simpl;
  intros x y vx vy Hrepx Hrepy.

intros _.
rewrite Rmult_1_r.
assert (U1 := Srepresents_interval _ _ Hrepx ).
assert (U2 := Srepresents_interval _ _ Hrepy ).
dec_tac; split; fourier.

destruct x as [x1 x']; destruct y as [y1 y'];
  inversion Hrepx; inversion Hrepy; subst.

replace (((IDR x1 + r) / Z2R b - (IDR y1 + r0) / Z2R b) * (Z2R b * Z2R b ^ n)) with 
   (((IDR x1 + r)  - (IDR y1 + r0) ) * Z2R b ^ n) by field_IDR.
compare_tac G1.
replace (IDR x1) with (IDR y1) by IDR_to_Z.
replace (IDR y1 + r - (IDR y1 + r0)) with 
  (r - r0) by ring.
apply IHn; assumption.

compare_tac G2;[|inversion 1|].

replace (IDR y1) with (IDR x1 + 2) by IDR_to_Z.
replace (IDR x1 + r - (IDR x1 + 2 + r0)) with 
  (r - (2 + r0)) by ring.
destruct x' as [x2 x''];
  inversion_to_variable x''; subst.
compare_eq_tac G3; [ |inversion 1].

destruct y' as [y2 y''];
  inversion_to_variable y''; subst.
compare_eq_tac G4; [ |inversion 1].
replace (IDR x2) with (Z2R b - 1) by IDR_to_Z.
replace (IDR y2) with (- Z2R b + 1) by IDR_to_Z.
replace ((Z2R b - 1 + r1) / Z2R b - (2 + (- Z2R b + 1 + r) / Z2R b)) with 
  ((IDR (- (1))%D + r1) / Z2R b - (IDR 1%D + r) / Z2R b) by field_IDR.
apply IHn;
  constructor; assumption || simpl_IDZ; omega.

replace (IDR y1) with (IDR x1 + 1) by IDR_to_Z.
replace (IDR x1 + r - (IDR x1 + 1 + r0)) with 
  (r - (1 + r0)) by ring.
destruct x' as [x2 x''];
  inversion_to_variable x''; subst.
compare_lt_tac G3; [inversion 1|].
destruct y' as [y2 y''];
  inversion_to_variable y''; subst.
compare_gt_tac G4; [inversion 1|].
replace (((IDR x2 + r1) / Z2R b - (1 + (IDR y2 + r) / Z2R b)) * Z2R b ^ n) with 
  (((IDR (x2-2)%D + r1) / Z2R b - ((IDR (y2 +Dmax-1)%D+ r) / Z2R b)) * Z2R b ^ n) by field_IDR.
apply IHn;
  constructor; assumption || simpl_IDZ; omega.

compare_tac G2;[| |inversion 1].

replace (IDR x1) with (IDR y1 + 2) by IDR_to_Z.
replace (IDR y1 + 2 + r - (IDR y1 + r0)) with 
  (r + 2 - r0) by ring.
destruct x' as [x2 x''];
  inversion_to_variable x''; subst.
compare_eq_tac G3; [ |inversion 1].

destruct y' as [y2 y''];
  inversion_to_variable y''; subst.
compare_eq_tac G4; [ |inversion 1].
replace (IDR y2) with (Z2R b - 1) by IDR_to_Z.
replace (IDR x2) with (- Z2R b + 1) by IDR_to_Z.
replace ((- Z2R b + 1 + r1) / Z2R b + 2 - (Z2R b - 1 + r) / Z2R b) with 
  ((IDR 1%D + r1) / Z2R b - (IDR (-(1))%D + r) / Z2R b) by field_IDR.
apply IHn;
  constructor; assumption || simpl_IDZ; omega.

replace (IDR x1) with (IDR y1 + 1) by IDR_to_Z.
replace (IDR y1 + 1 + r - (IDR y1 + r0)) with 
  (r +1 - r0) by ring.
destruct x' as [x2 x''];
  inversion_to_variable x''; subst.
compare_gt_tac G3; [inversion 1|].
destruct y' as [y2 y''];
  inversion_to_variable y''; subst.
compare_lt_tac G4; [inversion 1|].
replace (((IDR x2 + r1) / Z2R b + 1- (IDR y2 + r) / Z2R b) * Z2R b ^ n) with 
  (((IDR (x2+Dmax-1)%D + r1) / Z2R b - ((IDR (y2 -2)%D+ r) / Z2R b)) * Z2R b ^ n) by field_IDR.
apply IHn;
  constructor; assumption || simpl_IDZ; omega.
Qed.

Theorem exp_norm_correct :
  forall x y e x' y' vx vy,
    represents x vx ->
    represents y vy ->
    exp_norm x y = (e, x', y') ->
    represents (e, x') vx /\ represents (e, y') vy.
Proof.
intros x y e x' y' vx vy Hrepx Hrepy.
pose (Hbpos := base_positive).
assert (HbposR : 0 < Z2R b) by IDR_to_Z.

unfold exp_norm.
let_simpl; intros e1 x'' Hx;
let_simpl; intros e2 y'' Hy; subst.
elim (le_lt_dec e1 e2); intros G1;
  injection 1; intros; subst; intuition.

inversion Hrepx; subst.
replace (Z2R b ^ e1 * r) with (Z2R b ^e * (r / Z2R b ^ (e-e1))).
constructor.
apply str_add_zero_correct; assumption.
rewrite pow_minus; ZR0compat.
field; split; ZR0compat.

inversion Hrepy; subst.
replace (Z2R b ^ e2 * r) with (Z2R b ^e * (r / Z2R b ^ (e-e2))).
constructor.
apply str_add_zero_correct; assumption.
rewrite pow_minus; ZR0compat.
field; split; ZR0compat.
Qed.

Theorem compare_eq : 
  forall n x y vx vy, 
    represents x vx ->
    represents y vy ->
    compare x y n = Eq ->
    -2 <= (vx - vy) * Z2R b^ n<= 2.
Proof.
intros n x y vx vy Hrepx Hrepy.
assert (Hbpos := base_positive).
unfold compare.
let_simpl.
intros [e x'] y' Hexp.
generalize (exp_norm_correct _ _ _ _ _ _ _ Hrepx Hrepy Hexp). 
intros [H1 H2].
inversion H1; inversion H2; subst.
replace ((Z2R b ^ e * r - Z2R b ^ e * r0) * Z2R b ^ n) with 
  ((r - r0) * Z2R b ^ (n+e)) by (rewrite pow_add; ring).
apply Scompare_eq; assumption.
Qed.

Theorem compare_lt : 
  forall n x y vx vy, 
    represents x vx ->
    represents y vy ->
    compare x y n = Lt ->
    vx < vy.
Proof.
intros n x y vx vy Hrepx Hrepy.
assert (Hbpos := base_positive).
unfold compare.
let_simpl.
intros [e x'] y' Hexp.
generalize (exp_norm_correct _ _ _ _ _ _ _ Hrepx Hrepy Hexp). 
intros [H1 H2] Hcomp.
inversion H1; inversion H2; subst.
apply Rmult_lt_compat_l; [ZR0compat; IDR_to_Z| ].
generalize Hcomp; apply Scompare_lt; assumption.
Qed.

Theorem compare_gt : 
  forall n x y vx vy, 
    represents x vx ->
    represents y vy ->
    compare x y n = Gt ->
    vx > vy.
Proof.
intros n x y vx vy Hrepx Hrepy.
assert (Hbpos := base_positive).
unfold compare.
let_simpl.
intros [e x'] y' Hexp.
generalize (exp_norm_correct _ _ _ _ _ _ _ Hrepx Hrepy Hexp). 
intros [H1 H2] Hcomp.
inversion H1; inversion H2; subst.
apply Rlt_gt.
apply Rmult_lt_compat_l; [ZR0compat; IDR_to_Z| ].
assert (r > r0);[| fourier].
generalize Hcomp; apply Scompare_gt; assumption.
Qed.

Theorem compare2_eq : 
  forall n x y vx vy, 
    represents x vx ->
    represents y vy ->
    compare2 x y n = Eq ->
    -2 <= (vx - vy) * Z2R b^ n<= 2.
Proof.
intros n x y vx vy Hrepx Hrepy.
assert (Hbpos := base_positive).
unfold compare2.
let_simpl.
intros [e x'] y' Hexp.
generalize (exp_norm_correct _ _ _ _ _ _ _ Hrepx Hrepy Hexp). 
intros [H1 H2].
inversion H1; inversion H2; subst.
replace ((Z2R b ^ e * r - Z2R b ^ e * r0) * Z2R b ^ n) with 
  ((r - r0) * Z2R b ^ (n+e)) by (rewrite pow_add; ring).
apply Scompare2_eq; assumption.
Qed.

Theorem compare2_le : 
  forall n x y vx vy, 
    represents x vx ->
    represents y vy ->
    compare2 x y n = Lt ->
    vx <= vy.
Proof.
intros n x y vx vy Hrepx Hrepy.
assert (Hbpos := base_positive).
unfold compare2.
let_simpl.
intros [e x'] y' Hexp.
generalize (exp_norm_correct _ _ _ _ _ _ _ Hrepx Hrepy Hexp). 
intros [H1 H2] Hcomp.
inversion H1; inversion H2; subst.
apply Rmult_le_compat_l; [ZR0compat; IDR_to_Z| ].
generalize Hcomp; apply Scompare2_le; assumption.
Qed.

Theorem compare2_ge : 
  forall n x y vx vy, 
    represents x vx ->
    represents y vy ->
    compare2 x y n = Gt ->
    vx >= vy.
Proof.
intros n x y vx vy Hrepx Hrepy.
assert (Hbpos := base_positive).
unfold compare2.
let_simpl.
intros [e x'] y' Hexp.
generalize (exp_norm_correct _ _ _ _ _ _ _ Hrepx Hrepy Hexp). 
intros [H1 H2] Hcomp.
inversion H1; inversion H2; subst.
apply Rle_ge.
apply Rmult_le_compat_l; [ZR0compat; IDR_to_Z| ].
assert (r >= r0); [| fourier].
generalize Hcomp; apply Scompare2_ge; assumption.
Qed.

Theorem Smin_correct : 
  forall x y vx vy, 
    Srepresents x vx -> 
    Srepresents y vy -> 
    Srepresents (Smin x y) (Rmin vx vy).
Proof.
cofix.
intros x y vx vy Hrepx Hrepy.
str_unfold (Smin x y).
simpl.
destruct x as [x1 x']; destruct y as [y1 y'].
inversion Hrepx; inversion  Hrepy; subst.
rewrite Rmin_div_r; [| IDR_to_Z].
compare_tac G1.

replace (IDR y1) with (IDR x1) by IDR_to_Z.
rewrite Rmin_plus_l.
constructor; try assumption.
apply Smin_correct; assumption || simpl_IDZ; omega.
clear Smin_correct.
split;[ apply Rmin_le_l| apply Rmin_le_r]; intuition; fourier.

compare_eq_tac G2.

destruct x' as [x2 x''].
compare_tac G3.

destruct y' as [y2 y''].
compare_eq_tac G4.
replace (IDR y1 + r0) with (IDR x1 + (1 +r0));
  [| clear Smin_correct; (unfold IDR; rewrite <- G2; R_to_Z; ring)].
rewrite Rmin_plus_l.
constructor; [| | assumption].
apply Smin_correct; [assumption|].
clear Smin_correct.
inversion_to_variable y''; subst.
replace (1 + (IDR y2 + r1) / Z2R b) with 
  ((IDR 1%D + r1)/ Z2R b) by (simpl_IDR; rewrite G4; field_IDR).
constructor; assumption || simpl_IDZ; omega.
clear Smin_correct; split;[ apply Rmin_le_l| apply Rmin_le_r]; intuition; fourier.

clear Smin_correct.
rewrite Rmin_left ; [assumption|].
replace (IDR y1) with (IDR x1 + 1) by IDR_to_Z.
assert (r <= 1+ r0); [| dec_tac; fourier].
inversion_to_variable x''; inversion_to_variable y'';subst.
replace (1 + (IDR y2 + r2) / Z2R b) with 
  ((IDR y2 + Z2R b+ r2) / Z2R b) by field_IDR.
apply Rdiv_le_compat_r; [IDR_to_Z|].
replace (IDR x2) with 0 by IDR_to_Z.
assert (-Z2R b +2 <= IDR y2) by IDR_to_Z.
dec_tac; fourier.

clear Smin_correct.
rewrite Rmin_left ; [assumption|].
replace (IDR y1) with (IDR x1 + 1) by IDR_to_Z.
assert (r <= 0); [| dec_tac; fourier].
inversion_to_variable x''; subst.
apply Rle_div_r; [IDR_to_Z|].
assert (IDR x2 <= -1) by IDR_to_Z; dec_tac; fourier.

destruct y' as [y2 y''].
compare_gt_tac G4.

clear Smin_correct.
rewrite Rmin_left ; [assumption|].
replace (IDR y1) with (IDR x1 + 1) by IDR_to_Z.
assert (0 <= r0); [| dec_tac; fourier].
inversion_to_variable y''; subst.
apply Rle_div_l; [IDR_to_Z|].
assert (1 <= IDR y2) by IDR_to_Z; dec_tac; fourier.

replace (IDR x1 + r) with (IDR y1 + (-1 +r));
  [| clear Smin_correct; (unfold IDR; rewrite <- G2; R_to_Z; ring)].
rewrite Rmin_plus_l.
inversion_to_variable x''; subst.
constructor; [| | assumption].
apply Smin_correct; [|assumption].
clear Smin_correct.
replace (-1 + (IDR x2 + r1) / Z2R b) with 
  ((IDR (x2-Dbase)%D + r1)/ Z2R b) by field_IDR.
constructor; assumption || simpl_IDZ; omega.
clear Smin_correct; split;[ apply Rmin_le_l| apply Rmin_le_r]; intuition.
pattern (-1) at 1; replace (-1) with (-1+ 0) by ring.
apply Rplus_le_compat_l.
apply Rle_div_l; [IDR_to_Z|].
assert (1<= IDR x2) by IDR_to_Z; dec_tac; fourier.

clear Smin_correct.
rewrite Rmin_left ; [assumption|].
assert (IDR x1 <= IDR y1 -2) by IDR_to_Z; dec_tac; fourier.

compare_eq_tac G2.

destruct x' as [x2 x''].
compare_tac G3.

destruct y' as [y2 y''].
compare_eq_tac G4.
replace (IDR y1 + r0) with (IDR x1 + (-1 +r0));
  [| clear Smin_correct; (unfold IDR; rewrite G2; R_to_Z; ring)].
rewrite Rmin_plus_l.
inversion_to_variable y''; subst.
constructor; [| | assumption].
apply Smin_correct; [assumption|].
clear Smin_correct.
replace (-1 + (IDR y2 + r1) / Z2R b) with 
  ((IDR (-(1))%D + r1)/ Z2R b) by (simpl_IDR; rewrite G4; field_IDR).
constructor; assumption || simpl_IDZ; omega.
clear Smin_correct; split;[ apply Rmin_le_l| apply Rmin_le_r]; intuition.
pattern (-1) at 1; replace (-1) with (-1+ 0) by ring.
apply Rplus_le_compat_l.
apply Rle_div_l; [IDR_to_Z|].
assert (1<= IDR y2) by IDR_to_Z; dec_tac; fourier.

clear Smin_correct.
rewrite Rmin_right ; [assumption|].
replace (IDR y1) with (IDR x1 - 1) by IDR_to_Z.
assert (r0 <= 1+ r); [| dec_tac; fourier].
inversion_to_variable x''; inversion_to_variable y'';subst.
replace (1 + (IDR x2 + r1) / Z2R b) with 
  ((IDR x2 + Z2R b+ r1) / Z2R b) by field_IDR.
apply Rdiv_le_compat_r; [IDR_to_Z|].
replace (IDR x2) with 0 by IDR_to_Z.
assert (IDR y2 <= Z2R b -2) by IDR_to_Z.
dec_tac; fourier.

destruct y' as [y2 y''].
compare_lt_tac G4.

clear Smin_correct.
rewrite Rmin_right ; [assumption|].
replace (IDR x1) with (IDR y1 + 1) by IDR_to_Z.
assert (r0 <= 0); [| dec_tac; fourier].
inversion_to_variable y''; subst.
apply Rle_div_r; [IDR_to_Z|].
assert (IDR y2 <= -1) by IDR_to_Z; dec_tac; fourier.

replace (IDR x1 + r) with (IDR y1 + (1 +r));
  [| clear Smin_correct; (unfold IDR; rewrite G2; R_to_Z; ring)].
rewrite Rmin_plus_l.
constructor; [| | assumption].
apply Smin_correct; [|assumption].
clear Smin_correct.
inversion_to_variable x''; subst.
replace (1 + (IDR x2 + r1) / Z2R b) with 
  ((IDR (x2+Dbase)%D + r1)/ Z2R b) by field_IDR.
constructor; assumption || simpl_IDZ; omega.
clear Smin_correct; split;[ apply Rmin_le_l| apply Rmin_le_r]; intuition; fourier.

clear Smin_correct.
rewrite Rmin_right ; [assumption|].
replace (IDR x1) with (IDR y1 + 1) by IDR_to_Z.
assert (0 <= r); [| dec_tac; fourier].
inversion_to_variable x''; subst.
apply Rle_div_l; [IDR_to_Z|].
assert (1 <= IDR x2) by IDR_to_Z; dec_tac; fourier.

clear Smin_correct.
rewrite Rmin_right ; [assumption|].
assert (IDR y1 <= IDR x1 -2) by IDR_to_Z; dec_tac; fourier.
Qed.

Theorem min_correct : 
  forall x y vx vy, 
    represents x vx -> 
    represents y vy -> 
    represents (min x y) (Rmin vx vy).
Proof.
pose (Hbpos := base_positive).
intros x y vx vy Hrepx Hrepy.
unfold min, exp_norm.
inversion Hrepx; inversion Hrepy; subst.

elim le_lt_dec; intros.

replace (Z2R b^e*r) with (Z2R b^e0*(r/Z2R b^(e0-e)%nat)).
rewrite Rmin_mult_l.
constructor.
apply Smin_correct; try assumption.
apply str_add_zero_correct; try assumption.
unfold IDR; ZR0compat.
pattern e0 at 1; replace e0 with (e + (e0-e))%nat by omega. 
rewrite pow_add; field; unfold IDR; ZR0compat.

replace (Z2R b^e0*r0) with (Z2R b^e*(r0/Z2R b^(e-e0)%nat)).
rewrite Rmin_mult_l.
constructor.
apply Smin_correct; try assumption.
apply str_add_zero_correct; try assumption.
unfold IDR; ZR0compat.
pattern e at 1; replace e with (e0 + (e-e0))%nat by omega. 
rewrite pow_add; field; unfold IDR; ZR0compat.
Qed.

Theorem Smax_correct : 
  forall x y vx vy, 
    Srepresents x vx -> 
    Srepresents y vy -> 
    Srepresents (Smax x y) (Rmax vx vy).
Proof.
cofix.
intros x y vx vy Hrepx Hrepy.
str_unfold (Smax x y).
simpl.
destruct x as [x1 x']; destruct y as [y1 y'].
inversion Hrepx; inversion  Hrepy; subst.
rewrite Rmax_div_r; [| IDR_to_Z].
compare_tac G1.

replace (IDR y1) with (IDR x1) by IDR_to_Z.
rewrite Rmax_plus_l.
constructor; try assumption.
apply Smax_correct; assumption || simpl_IDZ; omega.
clear Smax_correct.
split;[ apply Rmax_le_l| apply Rmax_le_r]; intuition; fourier.

compare_eq_tac G2.

destruct y' as [y2 y''].
compare_tac G3.

destruct x' as [x2 x''].
compare_eq_tac G4.
replace (IDR x1 + r) with (IDR y1 + (-1 +r));
  [| clear Smax_correct; (unfold IDR; rewrite <- G2; R_to_Z; ring)].
rewrite Rmax_plus_l.
constructor; [| | assumption].
apply Smax_correct; [|assumption].
clear Smax_correct.
inversion_to_variable x''; subst.
replace (- 1 + (IDR x2 + r1) / Z2R b) with 
  ((IDR (-(1))%D + r1)/ Z2R b) by (simpl_IDR; rewrite G4; field_IDR).
constructor; assumption || simpl_IDZ; omega.
clear Smax_correct; split;[ apply Rmax_le_l| apply Rmax_le_r]; intuition; fourier.

clear Smax_correct.
rewrite Rmax_r; [assumption|].
replace (IDR y1) with (IDR x1 + 1) by IDR_to_Z.
assert (r <= 1+ r0); [| dec_tac; fourier].
inversion_to_variable x''; inversion_to_variable y'';subst.
replace (1 + (IDR y2 + r2) / Z2R b) with 
  ((IDR y2 + Z2R b+ r2) / Z2R b) by field_IDR.
apply Rdiv_le_compat_r; [IDR_to_Z|].
replace (IDR y2) with 0 by IDR_to_Z.
assert (IDR x2 <= Z2R b - 2) by IDR_to_Z.
dec_tac; fourier.

destruct x' as [x2 x''].
compare_lt_tac G4.

clear Smax_correct.
rewrite Rmax_r; [assumption|].
replace (IDR y1) with (IDR x1 + 1) by IDR_to_Z.
assert (0 <= -r ); [| dec_tac; fourier].
inversion_to_variable x''; subst.
rewrite <- Ropp_div_distr_l_reverse.
apply Rle_div_l; [IDR_to_Z|].
assert (IDR x2 <= -1) by IDR_to_Z; dec_tac; fourier.

replace (IDR y1 + r0) with (IDR x1 + (1 +r0));
  [| clear Smax_correct; (unfold IDR; rewrite <- G2; R_to_Z; ring)].
rewrite Rmax_plus_l.
inversion_to_variable y''; subst.
constructor; [| | assumption].
apply Smax_correct; [assumption|].
clear Smax_correct.
replace (1 + (IDR y2 + r1) / Z2R b) with 
  ((IDR (y2+Dbase)%D + r1)/ Z2R b) by field_IDR.
constructor; assumption || simpl_IDZ; omega.
clear Smax_correct; split;[ apply Rmax_le_l| apply Rmax_le_r]; intuition.
pattern (1) at 2; replace (1) with (1+ 0) by ring.
apply Rplus_le_compat_l.
apply Rle_div_r; [IDR_to_Z|].
assert (IDR y2 <= -1) by IDR_to_Z; dec_tac; fourier.

clear Smax_correct.
rewrite Rmax_r; [assumption|].
replace (IDR y1) with (IDR x1 + 1) by IDR_to_Z.
assert (0 <= r0); [| dec_tac; fourier].
inversion_to_variable y''; subst.
apply Rle_div_l; [IDR_to_Z|].
assert (1 <= IDR y2) by IDR_to_Z; dec_tac; fourier.

clear Smax_correct.
rewrite Rmax_r; [assumption|].
assert (IDR x1 <= IDR y1 -2) by IDR_to_Z; dec_tac; fourier.

compare_eq_tac G2.

destruct x' as [x2 x''].
compare_tac G3.

destruct y' as [y2 y''].
compare_eq_tac G4.
replace (IDR y1 + r0) with (IDR x1 + (-1 +r0));
  [| clear Smax_correct; (unfold IDR; rewrite G2; R_to_Z; ring)].
rewrite Rmax_plus_l.
inversion_to_variable y''; subst.
constructor; [| | assumption].
apply Smax_correct; [assumption|].
clear Smax_correct.
replace (-1 + (IDR y2 + r1) / Z2R b) with 
  ((IDR (-(1))%D + r1)/ Z2R b) by (simpl_IDR; rewrite G4; field_IDR).
constructor; assumption || simpl_IDZ; omega.
clear Smax_correct; split;[ apply Rmax_le_l| apply Rmax_le_r]; intuition.
pattern 1 at 2; replace (1) with (-1+ 2) by ring.
apply Rplus_le_compat_l.
apply Rle_div_r; [IDR_to_Z|].
replace (IDR y2) with (Z2R b -1)by IDR_to_Z.
assert (0 <= Z2R b) by R_to_Z.
dec_tac; fourier.

clear Smax_correct.
rewrite Rmax_l; [assumption|].
replace (IDR y1) with (IDR x1 - 1) by IDR_to_Z.
assert (r0 <= 1+ r); [| dec_tac; fourier].
inversion_to_variable x''; inversion_to_variable y'';subst.
replace (1 + (IDR x2 + r1) / Z2R b) with 
  ((IDR x2 + Z2R b+ r1) / Z2R b) by field_IDR.
apply Rdiv_le_compat_r; [IDR_to_Z|].
replace (IDR x2) with 0 by IDR_to_Z.
assert (IDR y2 <= Z2R b -2) by IDR_to_Z.
dec_tac; fourier.

destruct y' as [y2 y''].
compare_lt_tac G4.

clear Smax_correct.
rewrite Rmax_l; [assumption|].
replace (IDR x1) with (IDR y1 + 1) by IDR_to_Z.
assert (r0 <= 0); [| dec_tac; fourier].
inversion_to_variable y''; subst.
apply Rle_div_r; [IDR_to_Z|].
assert (IDR y2 <= -1) by IDR_to_Z; dec_tac; fourier.

replace (IDR x1 + r) with (IDR y1 + (1 +r));
  [| clear Smax_correct; (unfold IDR; rewrite G2; R_to_Z; ring)].
rewrite Rmax_plus_l.
inversion_to_variable x''; subst.
constructor; [| | assumption].
apply Smax_correct; [|assumption].
clear Smax_correct.
replace (1 + (IDR x2 + r1) / Z2R b) with 
  ((IDR (x2+Dbase)%D + r1)/ Z2R b) by field_IDR.
constructor; assumption || simpl_IDZ; omega.
clear Smax_correct; split;[ apply Rmax_le_l| apply Rmax_le_r]; intuition.
pattern 1 at 2; replace 1 with (1+ 0) by ring.
apply Rplus_le_compat_l.
apply Rle_div_r; [IDR_to_Z|].
assert (IDR x2 <= -1) by IDR_to_Z; dec_tac; fourier.

clear Smax_correct.
rewrite Rmax_l; [assumption|].
replace (IDR x1) with (IDR y1 + 1) by IDR_to_Z.
assert (0 <= r); [| dec_tac; fourier].
inversion_to_variable x''; subst.
apply Rle_div_l; [IDR_to_Z|].
assert (1 <= IDR x2) by IDR_to_Z; dec_tac; fourier.

clear Smax_correct.
rewrite Rmax_l; [assumption|].
assert (IDR y1 <= IDR x1 -2) by IDR_to_Z; dec_tac; fourier.
Qed.

Theorem max_correct : 
  forall x y vx vy, 
    represents x vx -> 
    represents y vy -> 
    represents (max x y) (Rmax vx vy).
Proof.
pose (Hbpos := base_positive).
intros x y vx vy Hrepx Hrepy.
unfold max, exp_norm.
inversion Hrepx; inversion Hrepy; subst.

elim le_lt_dec; intros.

replace (Z2R b^e*r) with (Z2R b^e0*(r/Z2R b^(e0-e)%nat)).
rewrite Rmax_mult_l.
constructor.
apply Smax_correct; try assumption.
apply str_add_zero_correct; try assumption.
unfold IDR; ZR0compat.
pattern e0 at 1; replace e0 with (e + (e0-e))%nat by omega. 
rewrite pow_add; field; unfold IDR; ZR0compat.

replace (Z2R b^e0*r0) with (Z2R b^e*(r0/Z2R b^(e-e0)%nat)).
rewrite Rmax_mult_l.
constructor.
apply Smax_correct; try assumption.
apply str_add_zero_correct; try assumption.
unfold IDR; ZR0compat.
pattern e at 1; replace e with (e0 + (e-e0))%nat by omega. 
rewrite pow_add; field; unfold IDR; ZR0compat.
Qed.

Theorem Sabs_correct : 
  forall x vx,
    Srepresents x vx -> 
    Srepresents (Sabs x) (Rabs vx).
Proof.  
cofix.
intros x vx Hrep.
str_unfold (Sabs x); simpl.
fold_decomp.
destruct x as [k x']; inversion Hrep; subst.

compare_tac G1.

replace (Rabs ((IDR k+r)/Z2R b)) with ((IDR 0%D+ Rabs r)/Z2R b).
constructor.
apply Sabs_correct; assumption.
clear Sabs_correct.
apply Rabs_le_le; assumption || fourier. 
simpl_IDZ; omega.
clear Sabs_correct.
replace (IDR k) with 0 by IDR_to_Z.
simpl_IDR; do 2 rewrite Rplus_0_l.
rewrite Rabs_div.
rewrite (Rabs_right (Z2R b)).
reflexivity.
unfold IDR; ZR0compat.
unfold IDR; ZR0compat.

clear Sabs_correct.
rewrite Rabs_right;[assumption |].
assert (1 <= IDR k) by IDR_to_Z.
apply Rle_ge; apply Rdiv_le_0_compat.
dec_tac; fourier.
IDR_to_Z.

clear Sabs_correct.
rewrite Rabs_left1.
apply Sopp_correct;assumption.
apply Rle_div_r; [IDR_to_Z | rewrite Rmult_0_l]. 
assert (IDR k <= -1) by IDR_to_Z; dec_tac; fourier.
Qed.

Theorem abs_correct : 
  forall x vx,  
    represents x vx ->
    represents (abs x) (Rabs vx).
Proof.
pose (Hbpos := base_positive).
intros x vx Hrepx.
unfold abs.
let_simpl; intros e x' H; subst.
inversion Hrepx; subst.
rewrite Rabs_mult.
rewrite (Rabs_right (Z2R b^e)).
constructor.
apply Sabs_correct; assumption.
ZR0compat; IDR_to_Z.
Qed.

End COMPARE_CORRECT.
