Require Import series_utils.
Require Export mult_correct.

Module COSINUS_CORRECT(Nums : NUMS).

Module Mult_correct := MULT_CORRECT(Nums).
Export Mult_correct.

Definition ch_term x n := 
 (x ^ (2*n)%nat / INR (fact (2*n)%nat))%R.

Definition cos_term_alt x := 
  tg_alt (ch_term x)%R.

Definition cos_term x n := 
 ( (-1) ^n * x ^ (2*n)%nat / INR (fact (2*n)%nat))%R.

Theorem ch_term_pos : 
  forall x n,
    0 <= ch_term x n.
Proof.
intros x n.
unfold ch_term.
apply Rdiv_le_0_compat.
rewrite pow_mult.
apply pow_le.
replace (x^2) with (Rsqr x) by (unfold Rsqr; ring).
apply Rle_0_sqr.
ZR0compat.
apply lt_O_fact.
Qed.

Theorem cos_term_alt_eq : 
  forall x n,
    cos_term_alt x n = cos_term x n.
Proof.  
intros x n.
unfold cos_term_alt, cos_term, ch_term, tg_alt, Rdiv; ring.
Qed.

Theorem cos_bounds_1 : 
  forall x s,
    infinit_sum (fun i => cos_term_alt x i) s ->
    Rabs s <= 1.
Proof.   
Admitted.

Theorem ch_term_decreasing :
  forall x, Un_decreasing (ch_term x).
Proof.
(*

unfold Un_decreasing.
intros p.
unfold ch_term.
replace ( x ^ (2 * p) / INR (fact (2 * p))) with 
  (( x ^ (2 * p) / INR (fact (2 * p)))*1) by ring.
replace (x ^ (2 * S p) / INR (fact (2 * S p))) with 
  (( x ^ (2 * p) / INR (fact (2 * p))) * (x^2/INR ((2*p+1)*(2*(p+1))))).
apply Rmult_le_compat_l.
generalize (ch_term_pos x p); unfold ch_term; tauto.
apply Rle_div_r.
ZR0compat.
replace ((2 * p + 1) * (2 * (p + 1)))%nat with 
  (S (1+6*p+4*p*p))%nat by ring.
apply lt_O_Sn.


*)
Admitted.

Theorem ch_term_cv_0 : 
  forall x, Un_cv (ch_term x) 0.
Proof.
Admitted.

Theorem cos_bounds : 
  forall n x s,
    Rabs x <= 1 ->
    infinit_sum (fun i => cos_term_alt x (i+ n)) s ->
    Rabs s <= Rabs x ^ (2*n) / INR (fact (2*n)).
Proof.
intros n x s H1 H2.
destruct n.
  simpl.
  replace (1/1) with 1 by field. 
  apply cos_bounds_1 with x.
  apply infinit_sum_ext with (2:= H2).
  intros n; rewrite plus_0_r; ring.
 
apply Rle_trans with (ch_term x (S n)).
apply (infinit_sum_tg_alt).
  apply ch_term_decreasing.
  apply ch_term_cv_0.
  unfold cos_term_alt in H2; assumption.

unfold ch_term.
apply Rdiv_le_compat_r.
ZR0compat; apply lt_O_fact.
rewrite RPow_abs.
apply RRle_abs.
Qed.













End COSINUS_CORRECT.