Require Export proof_utils.

(*Interface of types of digits (not already modular arithmetic)
   and big numbers used in intermediate calculations *)
Module Type NUMS.
  
  (* DIGITS TYPE *)
  
  Parameter b : Z.
  Parameter digit : Set.
  Parameter D0 D1 Dbase Dmax : digit.  
  Parameter Dplus Dminus Dmult : digit -> digit -> digit.
  Parameter Ddiv_eucl : digit -> digit -> digit*digit.
  Parameter Dopp Dlog2: digit -> digit.
  Parameter Dcompare : digit -> digit -> comparison.
  Parameter IDN : digit -> nat.
  
  Delimit Scope digit_scope with D.

  Infix "+" := Dplus : digit_scope.
  Infix "-" := Dminus : digit_scope.
  Infix "*" := Dmult : digit_scope.
  Infix "?=" := Dcompare : digit_scope.
  Notation "- x" := (Dopp x) : digit_scope.
  Notation "0" := D0 : digit_scope.
  Notation "1" := D1 : digit_scope.
  Notation "2" := (1+1)%D : digit_scope.
  Notation "4" := (2+2)%D : digit_scope.

  Axiom base_positive : (4 < b)%Z. 

  Axiom IDZ : digit -> Z.
  Axiom D0_IDZ : IDZ 0%D = 0%Z.
  Axiom D1_IDZ : IDZ 1%D = 1%Z.
  Axiom Dbase_IDZ : IDZ Dbase = b.
  Axiom Dmax_IDZ : IDZ Dmax = (b -1)%Z.
  Axiom plus_IDZ: forall m n : digit, 
    IDZ (m + n)%D = (IDZ m + IDZ n)%Z.
  Axiom minus_IDZ: forall m n : digit, 
    IDZ (m - n)%D = (IDZ m - IDZ n)%Z.
  Axiom mult_IDZ: forall m n : digit, 
    IDZ (m * n)%D = (IDZ m * IDZ n)%Z.
  Axiom div_eucl_IDZ : forall m n, 
    let (q, r) := Ddiv_eucl m n in 
      (IDZ q, IDZ r) = Zdiv_eucl (IDZ m) (IDZ n).
  Axiom opp_IDZ: forall n, IDZ (- n)%D = (- IDZ n)%Z.
  Axiom log_IDZ : forall n : digit, IDZ (Dlog2 n) = N_digits (IDZ n). 
  Axiom compare_IDZ:
    forall x y : digit,
      match (x ?= y)%D with
	| Eq => IDZ x = IDZ y
	| Lt => (IDZ x < IDZ y)%Z
	| Gt => (IDZ x > IDZ y)%Z
      end.
  Axiom IDZ_IDN : forall n, IDN n = Zabs_nat (IDZ n). 

  
(*
  Parameter t_ : Set.
  Parameter Dlow Dhigh : t_ -> digit.

  Parameter Dplus_ : digit -> digit -> t_.




  Axiom plus_IDZ_0 : 
    forall m n, 
      (-b < IDZ m + IDZ n < b ->
	IDZ (Dlow (Dplus_ m n)) = IDZ m + IDZ n /\ 
	Dhigh (Dplus_ m n) = D0)%Z.

  Axiom plus_IDZ_1 : 
    forall m n, 
      (b <= IDZ m + IDZ n <= 2*b-2 ->
	IDZ (Dlow (Dplus_ m n)) = IDZ m + IDZ n -b /\ 
	Dhigh (Dplus_ m n) = D1)%Z.

  Axiom plus_IDZ_m1 : 
    forall m n, 
      (-2*b+2 <= IDZ m + IDZ n <= - b ->
	IDZ (Dlow (Dplus_ m n)) = IDZ m + IDZ n + b/\ 
	Dhigh (Dplus_ m n) = (-D1)%D)%Z.
*)
  (* BIG NUMBERS TYPE *)

  Parameter big : Set.
  Parameter B0 B1 Bbase: big.
  Parameter Bplus Bmult Bminus Bpower : big -> big -> big. 
  Parameter Bdiv_eucl : big -> big -> big*big.
  Parameter Bopp : big -> big.
  Parameter Ble_bool : big -> big -> bool.
  
  Delimit Scope big_scope with B.

  Infix "+" := Bplus : big_scope.
  Infix "-" := Bminus : big_scope.
  Infix "*" := Bmult : big_scope.
  Infix "^" := Bpower : big_scope.
  Infix "<=" := Ble_bool : big_scope.
  Notation "- x" := (Bopp x) : big_scope.
  Notation "0" := B0 : big_scope.
  Notation "1" := B1 : big_scope.
  Notation "2" := (1+1)%B : big_scope.
  Notation "3" := (2+1)%B : big_scope.
  Notation "4" := (2+2)%B : big_scope.
  Notation "8" := (4+4)%B : big_scope.
  Notation "9" := (3*3)%B : big_scope.
  Notation "16" := (8+8)%B : big_scope.
  Notation "18" := (16+2)%B : big_scope.
  Notation "27" := (9*3)%B : big_scope.
  Notation "81" := (9*9)%B : big_scope.
  Notation "1296" := (16*81)%B : big_scope.
  
  Axiom IBZ : big -> Z.
  Axiom IZB : Z -> big.
  Axiom Bbase_IBZ : IBZ Bbase = b.
  Axiom IZB_IBZ_eq : forall x c, IZB x = c <-> x = IBZ c. 
  Axiom B0_IBZ : IBZ 0%B = 0%Z.
  Axiom B1_IBZ : IBZ 1%B = 1%Z.
  Axiom plus_IBZ: forall m n : big, IBZ (m + n)%B = (IBZ m + IBZ n)%Z.
  Axiom mult_IBZ: forall m n : big, IBZ (m * n)%B = (IBZ m * IBZ n)%Z.
  Axiom minus_IBZ: forall m n : big, IBZ (m - n)%B = (IBZ m - IBZ n)%Z.
  Axiom power_IBZ: forall m n : big, IBZ (m ^ n)%B = (IBZ m ^ IBZ n)%Z.
  Axiom div_IBZ : forall m n, let (k,r) := Bdiv_eucl m n in 
    IBZ k = (IBZ m/IBZ n)%Z.
  Axiom mod_IBZ : forall m n, let (k,r) := Bdiv_eucl m n in 
    IBZ r = (IBZ m mod IBZ n)%Z.  
  Axiom opp_IBZ: forall n, IBZ (- n)%B = (- IBZ n)%Z.
  Axiom le_IBZ: forall m n : big, (m <= n)%B = (Zle_bool (IBZ m) (IBZ n)).

  (* CONVERSIONS BETWEEN DIGIT AND BIG *)
  
  Parameter digit_of_big : big -> digit.
  Parameter big_of_digit : digit -> big.
  
  Axiom digit_of_big_of_digit_inv : forall d, digit_of_big (big_of_digit d) = d.
  Axiom big_of_digit_of_big_inv : forall b, big_of_digit (digit_of_big b) = b.
  Axiom IBZ_IDZ : forall d, IBZ (big_of_digit d) = IDZ d.

End NUMS.

(* Extension of interface with injection in R*)
Module NUMS_EXT (Nums : NUMS).

Export Nums.

Theorem IBZ_IZB : forall x, IBZ (IZB x) = x.
Proof.
intros x.
symmetry.
generalize (IZB_IBZ_eq x (IZB x)).
tauto.
Qed.

Theorem IZB_IBZ : forall x, IZB (IBZ x) = x.
Proof.
intros x.
generalize (IZB_IBZ_eq (IBZ x) x).
tauto.
Qed.

Theorem IDZ_IBZ : forall x : big, IDZ (digit_of_big x) = IBZ x.
Proof.
intros x.
pattern x at 2; rewrite <- (big_of_digit_of_big_inv x).
rewrite IBZ_IDZ; auto.
Qed.

Ltac simpl_IDZ :=
  repeat (rewrite plus_IDZ || rewrite minus_IDZ || rewrite opp_IDZ ||
    rewrite mult_IDZ || rewrite log_IDZ ||
      rewrite digit_of_big_of_digit_inv || rewrite IDZ_IBZ ||
	rewrite D0_IDZ || rewrite D1_IDZ || rewrite Dmax_IDZ || 
	  rewrite IDZ_IDN || rewrite Dbase_IDZ). 

Ltac fold_IDZ :=
  repeat (rewrite <- plus_IDZ || rewrite <- minus_IDZ ||
    rewrite <- opp_IDZ || rewrite <- mult_IDZ ||
       rewrite <- log_IDZ || rewrite <- IDZ_IDN 
	 || rewrite <- D0_IDZ || rewrite <- D1_IDZ || rewrite <- Dbase_IDZ). 

Ltac simpl_IBZ :=
  repeat (rewrite plus_IBZ || rewrite minus_IBZ ||
    rewrite opp_IBZ || rewrite mult_IBZ || rewrite le_IBZ || 
      rewrite big_of_digit_of_big_inv || rewrite IBZ_IDZ ||
	rewrite power_IBZ || rewrite B0_IBZ|| rewrite B1_IBZ ||
	  rewrite IBZ_IZB || rewrite IZB_IBZ ||
	    rewrite Bbase_IBZ).

Ltac fold_IBZ :=
  repeat (rewrite <- plus_IBZ || rewrite <- minus_IBZ ||
    rewrite <- opp_IBZ || rewrite <- mult_IBZ ||  
      rewrite <- power_IBZ || rewrite <- B0_IBZ|| rewrite <- B1_IBZ ||
	rewrite <- Bbase_IBZ).

Definition IDR (d: digit) : R := Z2R (IDZ d).
Definition IBR (b: big) : R := Z2R (IBZ b).

Ltac simpl_IDR := 
  unfold IDR; unfold IBR; simpl_IDZ; simpl_Z2R. 

Ltac field_IDR := 
  simpl_IDR; field_Z.

Ltac IDR_to_Z := 
  simpl_IDR; R_to_Z.

Ltac fold_IDR :=
  match goal with 
    |- context [ Z2R (IDZ ?k)] => 
      change (Z2R (IDZ k)) with (IDR k); fold_IDR
    | _ => idtac
  end.

Ltac simpl_IBR := 
  unfold IDR; unfold IBR; simpl_IBZ; simpl_Z2R. 

Ltac field_IBR := 
  simpl_IBR; field_Z.

Ltac IBR_to_Z := 
  simpl_IBR; R_to_Z.

Ltac fold_IBR :=
  match goal with 
    |- context [ Z2R (IBZ ?k)] => 
      change (Z2R (IBZ k)) with (IBR k); fold_IBR
    | _ => idtac
  end.

End NUMS_EXT.
