Require Export mult_digit_correct.

Module MAKE_DIGIT_CORRECT(Nums : NUMS).
Module Mult_digit_correct := MULT_DIGIT_CORRECT(Nums).
Export Mult_digit_correct.

Theorem make_digit_eq : 
  forall x vx, 
    Srepresents x vx -> 
    Srepresents (make_digit x) vx.
Proof.
intros x vx Hrep.
unfold make_digit.
destruct x as [k1 [k2 x]].
unfold Dle_bool;
compare_gt_tac G1; compare_gt_tac G2; try compare_gt_tac G3;
  try assumption; inversion_to_variable x; subst;
    (rep_constr; [idtac| field_IDR]).

constructor.
constructor; (assumption || simpl_IDZ; omega).
apply Rle_div_1; IDR_to_Z.
fold_IDR; dec_tac; split.
apply Rle_trans with 0%R.
IDR_to_Z.
assert (-Z2R b+1 <= IDR k2)%R; [IDR_to_Z| fourier].
assert (IDR k2<= -1)%R; [IDR_to_Z| fourier].
simpl_IDZ; omega.

constructor.
constructor; (assumption || simpl_IDZ; omega).
apply Rle_div_1; IDR_to_Z.
fold_IDR; dec_tac; split.
assert (1<= IDR k2)%R; [IDR_to_Z| fourier].
apply Rle_trans with 0%R.
assert (IDR k2 <= Z2R b-1)%R; [IDR_to_Z| fourier].
IDR_to_Z.
simpl_IDZ; omega.
Qed.

Fixpoint make_k_digits (x y : Stream digit) (n : nat) {struct n} :=
  match n with 
    | O => make_digit y
    | S n => 
      match x with 
        | x1::x' =>
          x1::make_k_digits x'
          (match y with 
             y1::y' => 
             match x1?= y1 with
               | Eq => y'    
               | Gt => 
                 match y' with 
                   | y2::y'' => (y2 - Dbase)::y''
                 end
               | Lt => 
                 match y' with 
                   | y2::y'' => (y2 + Dbase)::y''
                 end
             end
           end)%D n 
      end
  end.

Fixpoint truncate x n {struct n} :=
  match n with 
    | 0 => Szero
    | S n => 
      match x with 
        | d::x => d::truncate x n
      end
  end.


Theorem Srep_truncate_head :
  forall k x vx, 
    Srepresents (truncate x k) vx <->
    Srep_head x k vx.
Proof.
intros k; induction k; intros x vx; simpl.
split; intros Hrep.
rewrite <- Srep_Szero_eq_0; [constructor | assumption].
inversion Hrep; apply rep_Szero.

destruct x as [x1 x'];split; intros Hrep;
inversion Hrep; subst; constructor; try assumption.
rewrite <- IHk; assumption.
rewrite IHk; assumption.
rabs_intros; apply (Srep_head_inter_1 k x'); assumption.
Qed.


Theorem Rabs_add_lt_r:
  forall w x y z : R, 
    Rabs w <= y ->
    Rabs x < z -> 
    Rabs (w + x) < y + z.
Proof.
intros w x y z Hwy Hxz.
apply Rle_lt_trans with (1:= (Rabs_triang w x)).
fourier.
Qed.

Theorem Rabs_add_lt_l:
  forall w x y z : R, 
    Rabs w < y ->
    Rabs x <= z -> 
    Rabs (w + x) < y + z.
Proof.
intros w x y z Hwy Hxz.
apply Rle_lt_trans with (1:= (Rabs_triang w x)).
fourier.
Qed.

Theorem  make_k_digit_tech1_old : 
  forall k x1 x vx vy,
    Srep_head (x1::x) (S k) vx ->
    Rabs (vx - vy) < 1 / Z2R b ^ S k ->
    Rabs (IDR x1 / Z2R b- vy ) < 1/Z2R b.
Proof.    
intros k x1 x vx vy Hrepx Hxy.
inversion Hrepx; subst.
replace (IDR x1 / Z2R b - vy) with 
  (((IDR x1 + vx0)/ Z2R b - vy) + - vx0 / Z2R b) by (unfold Rdiv; ring).
replace (1/Z2R b) with (1 / Z2R b ^ S k + ( 1 - 1 / Z2R b^k)/ Z2R b)
  by (simpl; field_Z; ZR0compat).
apply Rabs_add_lt_l; [assumption| ].
rewrite Rabs_div; [| ZR0compat].
rewrite Rabs_Ropp.
rewrite (Rabs_right (Z2R b));[| ZR0compat].
apply Rdiv_le_compat_r;[ZR0compat|].
apply Srep_head_inter with x; assumption.
Qed.

Theorem  make_k_digit_tech2_old : 
  forall k x1 x vx y1 y vy,
    Srep_head (x1::x) (S k) vx ->
    Srepresents (y1::y) vy ->
    Rabs (vx - vy) < 1 / Z2R b ^ S k ->
    (-1 <= IDZ x1 - IDZ y1 <= 1)%Z.
Proof.    
intros k x1 x vx y1 y vy Hrepx Hrepy Hxy.
assert ( U1:    Rabs (IDR x1 / Z2R b- vy ) < 1/Z2R b).
  apply make_k_digit_tech1_old with (1:= Hrepx); assumption.
inversion Hrepy; subst.  
assert (U2 : Rabs (IDR x1 - IDR y1) < 2).
replace (IDR x1 - IDR y1) with ((IDR x1 - (IDR y1+r)) + r) by ring.
apply Rabs_add_lt_l ;[| rabs_elim; assumption].
replace (Rabs (IDR x1 - (IDR y1 + r))) with 
  (((Rabs (IDR x1 /Z2R b - (IDR y1 + r)/ Z2R b)))* Z2R b).
apply Rdiv_lt_l; [R_to_Z|assumption].
replace (IDR x1 / Z2R b - (IDR y1 + r) / Z2R b) with 
  ((IDR x1 - (IDR y1 + r)) / Z2R b) by (unfold Rdiv; ring).
rewrite Rabs_div; [| R_to_Z].
rewrite (Rabs_right (Z2R b)); [field_Z| ZR0compat].
assert (-2 < IDZ x1 - IDZ y1 < 2)%Z; [| omega].
destruct (Rabs_def2 _ _ U2) as [U3 U4].
split; apply lt_Z2R; simpl_Z2R; fold_IDR; assumption.
Qed.

Theorem  make_k_digit_tech : 
  forall k xh x vx yh y vy,
    Srep_head x k vx ->
    Srepresents y vy -> 
    Z2R b ^ k * Rabs (( xh + vx) - ( yh + vy)) < 1  ->
    Rabs (xh - ( yh + vy )) < 1.
Proof. 
assert (Hbpos:=base_positive).   
intros k xh x vx yh y vy Hrepx Hrepy Hxy.
replace (xh - (yh + vy)) with 
  (((xh + vx) - (yh + vy)) + - vx) by ring.
replace 1 with (1 / Z2R b ^ k + ( 1 - 1 / Z2R b^k))
  by ring.
apply Rabs_add_lt_l.
  apply Rlt_div_l; [ZR0compat|rewrite Rmult_comm; assumption].
rewrite Rabs_Ropp.
apply Srep_head_inter with x; assumption.
Qed.

Theorem  make_k_digit_tech1 : 
  forall k x1 x vx y1 y vy,
    Srep_head x k vx ->
    Srepresents y vy -> 
    Z2R b ^ k * Rabs ((IDR x1 + vx) - (IDR y1 + vy)) < 1  ->
    Rabs (IDR x1 - (IDR y1 + vy )) < 1.
Proof. 
assert (Hbpos:=base_positive).  
intros k x1 x vx y1 y vy.
apply make_k_digit_tech.
Qed.

Theorem  make_k_digit_tech2 : 
  forall k x1 x2 x vx y1 y2 y vy,
    Srep_head x k vx ->
    Srepresents y vy -> 
    Z2R b ^ k * Rabs ((Z2R b* IDR x1 + IDR x2 + vx) - (Z2R b * IDR y1 + IDR y2 + vy)) < 1  ->
    Rabs (Z2R b * IDR x1 +IDR x2 - (Z2R b* IDR y1 + IDR y2 + vy )) < 1.
Proof. 
assert (Hbpos:=base_positive).   
intros k x1 x2 x vx y1 y2 y vy.
apply make_k_digit_tech.
Qed.

Theorem  make_k_digit_tech3 : 
  forall k x1 x vx y1 y vy,
    Srep_head x k vx ->
    Srepresents y vy -> 
    Z2R b ^ k * Rabs ((IDR x1 + vx) - (IDR y1 + vy)) < 1 ->
    (-1 <= IDZ x1 - IDZ y1 <= 1)%Z.
Proof.    
intros k x1 x vx y1 y vy Hrepx Hrepy Hxy.
assert ( U1:    Rabs (IDR x1 - (IDR y1 + vy )) < 1).
  apply make_k_digit_tech1 with (1:= Hrepx) (2:= Hrepy); assumption.

assert (U2 : Rabs (IDR x1 - IDR y1) < 2).
replace (IDR x1 - IDR y1) with ((IDR x1 - (IDR y1+vy)) + vy) by ring.
apply Rabs_add_lt_l; [assumption|].
rabs_elim; apply Srepresents_interval with y; assumption.

assert (-2 < IDZ x1 - IDZ y1 < 2)%Z; [| omega].
destruct (Rabs_def2 _ _ U2) as [U3 U4].
split; apply lt_Z2R; simpl_Z2R; fold_IDR; assumption.
Qed.

Theorem Str_nth_S : 
  forall A n x (k : A), 
    Str_nth (S n) (k::x) =
    Str_nth n x. 
Proof.
intros n x k H.
unfold Str_nth in *; reflexivity.
Qed.

Theorem Str_nth_0 : 
  forall A x (k: A), 
    Str_nth 0 (k::x) = k.
Proof.
intros A x k.
unfold Str_nth; reflexivity.
Qed.

Theorem make_digit_Str_nth_0 :
  forall x vx,
    Srepresents x vx -> 
    (- b < IDZ (Str_nth 0 (make_digit x)) < b)%Z.
Proof.
intros x vx Hrepx.
assert (Hrepmdx := make_digit_eq x vx Hrepx).
inversion Hrepmdx.
rewrite Str_nth_0; assumption.
Qed.

Theorem Srep_Cons : 
  forall x vx d, 
    -1 <= vx <= 1 -> 
    (-b < IDZ d < b)%Z -> 
    (Srepresents (d::x) ((IDR d + vx) / Z2R b) <-> Srepresents x vx).
Proof.
intros x vx d Hvx Hd; split; intros Hrep.
inversion Hrep; subst.
replace vx with r ;[assumption|].
replace r with (((IDR d + r) / Z2R b) * Z2R b - IDR d) by field_IDR.
rewrite H1; field_IDR.

constructor; assumption.
Qed.

Theorem make_k_digits_eq : 
  forall k x y vx vy,
    Srep_head x k vx ->
    Srepresents y vy -> 
    (forall n, (n <= k)%nat -> 
      (IDZ (Str_nth n x) <= b - 2 ->
	IDZ (Str_nth (S n) x) <= b - 2)%Z) ->  
    (forall n, (n <= k)%nat -> 
      (-b +2 <= IDZ (Str_nth n x) ->
	-b+2 <= IDZ (Str_nth (S n) x))%Z) ->  
    Z2R b ^ S k * Rabs (vx - vy) < Z2R b - 1 ->
    Srepresents (make_k_digits x y k) vy /\   
    (- b < IDZ (Str_nth k (make_k_digits x y k)) < b)%Z.
Proof.
assert (Hbpos:= base_positive).
intros k; induction k; intros x y vx vy Hx Hy Hd1 Hd2 Hxy; simpl.
split;[ 
  apply make_digit_eq; assumption
  | apply make_digit_Str_nth_0 with vy; assumption].

destruct x as [x1 x']; destruct y as [y1 y'];
inversion Hy; inversion Hx; subst.
replace ((IDR x1 + vx0) / Z2R b - (IDR y1 + r) / Z2R b) with 
  (((IDR x1 + vx0) - (IDR y1 + r)) / Z2R b) in Hxy by (unfold Rdiv; ring).
replace (Z2R b ^ S (S k) * Rabs (((IDR x1 + vx0) - (IDR y1 + r)) / Z2R b)) with 
  (Z2R b ^ S k * Rabs ((IDR x1 + vx0) - (IDR y1 + r))) in Hxy; [|
    rewrite Rabs_div; [|ZR0compat]; rewrite (Rabs_right (Z2R b));[|ZR0compat];
      simpl; field_Z].

assert( U1 : Z2R b ^ k * Rabs (IDR x1 + vx0 - (IDR y1 + r)) < 1).
  apply Rmult_lt_reg_l with (Z2R b);[ZR0compat|].
  apply Rlt_trans with (Z2R b -1);[| R_to_Z].
  apply Rle_lt_trans with (2:= Hxy); right; simpl; ring.

assert (U2 :  Rabs (IDR x1 - (IDR y1 + r )) < 1).
  apply make_k_digit_tech1 with (1:= H8) (2:= H1); assumption.

assert (U3 : (-1 <= IDZ x1 - IDZ y1 <= 1)%Z).
  apply make_k_digit_tech3 with (1:=H8) (2:= H1); assumption.

assert (Hd1' : forall n : nat,
   (n <= k)%nat ->
   (IDZ (Str_nth n x') <= b - 2)%Z -> (IDZ (Str_nth (S n) x') <= b - 2)%Z).
  intros n Hn Hnx.
  apply (Hd1 (S n)); [omega|].
  unfold Str_nth in *; simpl; assumption.

assert (Hd2' : forall n : nat,
   (n <= k)%nat ->
   (-b+2 <= IDZ (Str_nth n x'))%Z -> (-b+2 <= IDZ (Str_nth (S n) x'))%Z).
  intros n Hn Hnx.
  apply (Hd2 (S n)); [omega|].
  unfold Str_nth in *; simpl; assumption.

compare_tac G1.
(* case x1 = y1 *)
clear Hd1 Hd2.
replace (IDR y1) with (IDR x1) in * by (IDR_to_Z).
rewrite Str_nth_S.
rewrite Srep_Cons; try assumption.
apply IHk with vx0; try assumption.

apply Rle_lt_trans with (2:=Hxy); right.
replace ((IDR x1 + vx0) - (IDR x1 + r)) with  (vx0 - r); ring.

(* case  x1 < y1 => y1 = x1+1 *)
clear Hd2.
assert (R1 : IDR y1 = IDR (x1+1)%D) by IDR_to_Z.
clear U3; rewrite R1 in *.
replace (IDR x1 - (IDR (x1 + 1)%D + r)) with (-(1+ r)) in U2 by (simpl_IDR; ring).
rewrite Rabs_Ropp in U2.
replace (IDR x1 +vx0 - (IDR (x1 + 1)%D + r)) with (vx0 - (1+ r)) in Hxy by (simpl_IDR; ring).

destruct y' as [y2 y''].
replace (IDR (x1 + 1)%D + r) with (IDR x1 + (1+r)) by (simpl_IDR; ring).

rewrite Str_nth_S.
rewrite Srep_Cons; [rabs_intros;intuition | assumption|].
inversion H1; subst.
replace (1 + (IDR y2 + r0) / Z2R b)
  with ((IDR (y2+Dbase)%D + r0)/ Z2R b) by field_IDR.
apply IHk with vx0; try assumption; clear IHk Hd1' Hd2'.
constructor; try assumption.
simpl_IDZ; assert (IDZ y2 < 0)%Z;[| omega].
destruct k.
  inversion H8; subst.
  apply lt_Z2R; simpl_Z2R; fold_IDR.
  assert (Z2R b+ IDR y2 + r0 < Z2R b-1);[| dec_tac; fourier].
  apply Rle_lt_trans with (2:= Hxy); right.
  rewrite Rminus_0_l; rewrite Rabs_Ropp.
  rewrite Rabs_right; [simpl; field_Z|].
  replace ( 1 + (IDR y2 + r0) / Z2R b) with 
    ((Z2R b + IDR y2 + r0)/ Z2R b) by field_Z.
  apply Rle_ge; apply Rdiv_le_0_compat;[| ZR0compat].
  assert (- Z2R b + 1 <= IDR y2) by IDR_to_Z.
  dec_tac; fourier.
  
destruct x' as [x2 x'']; inversion H8; subst.
assert(U3: Rabs (Z2R b * IDR x1 +IDR x2 - (Z2R b* IDR y1+ IDR y2 + r0 )) < 1).
  apply make_k_digit_tech2 with (1:=H9) (2:=H3).
  apply Rmult_lt_reg_l with (Z2R b); [ZR0compat|].
  apply Rlt_trans with (Z2R b - 1); [| R_to_Z].
  apply Rle_lt_trans with (2:=Hxy); right.
  replace (Z2R b * IDR x1 + IDR x2 + vx - (Z2R b * IDR y1 + IDR y2 + r0)) 
    with (Z2R b * ((IDR x2 + vx) / Z2R b - (1 + (IDR y2 + r0) / Z2R b)))
    by (rewrite R1; field_IDR).
  rewrite Rabs_mult; rewrite (Rabs_right (Z2R b)); [simpl; ring|ZR0compat].

replace (Z2R b * IDR x1 + IDR x2 - (Z2R b * IDR y1 + IDR y2 + r0))
  with  (IDR x2 - (Z2R b  + IDR y2 + r0)) in U3
  by (rewrite R1; simpl_IDR; ring). 
rabs_elim.
apply lt_Z2R; simpl_Z2R; fold_IDR.
apply Rlt_le_trans with (1 + IDR x2 - Z2R b - r0); [dec_tac; fourier|].
assert (IDR x2 <= Z2R b - 2); [| dec_tac; fourier].
IDR_to_Z.
generalize (Hd1 O); unfold Str_nth; simpl.
intros U4; apply U4; omega.

replace ((IDR (y2 + Dbase)%D + r0) / Z2R b) with 
  (1 + (IDR y2 + r0) / Z2R b) by field_IDR; assumption.

(*case x1 > y1 => x1 = y1 + 1*)
clear Hd1.
assert (R1 : IDR y1 = IDR (x1-1)%D) by IDR_to_Z.
clear U3; rewrite R1 in *.
replace (IDR x1 - (IDR (x1 - 1)%D + r)) with (-(- 1+ r)) in U2 by (simpl_IDR; ring).
rewrite Rabs_Ropp in U2.
replace (IDR x1 +vx0 - (IDR (x1 - 1)%D + r)) with (vx0 - (- 1+ r)) in Hxy by (simpl_IDR; ring).

destruct y' as [y2 y''].
replace (IDR (x1 - 1)%D + r) with (IDR x1 + (-1+r)) by (simpl_IDR; ring).

rewrite Str_nth_S.
rewrite Srep_Cons; [rabs_intros;intuition | assumption|].
inversion H1; subst.
replace (- 1 + (IDR y2 + r0) / Z2R b)
  with ((IDR (y2-Dbase)%D + r0)/ Z2R b) by field_IDR.
apply IHk with vx0; try assumption; clear IHk Hd1' Hd2'.
constructor; try assumption.
simpl_IDZ; assert (- IDZ y2 < 0)%Z;[| omega].
destruct k.
  inversion H8; subst.
  apply lt_Z2R; simpl_Z2R; fold_IDR.
  assert (Z2R b - IDR y2 - r0 < Z2R b-1);[| dec_tac; fourier].
  apply Rle_lt_trans with (2:= Hxy); right.
  rewrite Rminus_0_l.
  rewrite Rabs_right; [simpl; field_Z|].
  replace ( -1 + (IDR y2 + r0) / Z2R b) with 
    ((-Z2R b + IDR y2 + r0)/ Z2R b) by field_Z.
  replace  (- ((- Z2R b + IDR y2 + r0) / Z2R b)) with 
     ((Z2R b - IDR y2 - r0) / Z2R b) by field_Z.
  apply Rle_ge; apply Rdiv_le_0_compat;[| ZR0compat].
  assert (IDR y2 <= Z2R b -1) by IDR_to_Z.
  dec_tac; fourier.
  
destruct x' as [x2 x'']; inversion H8; subst.
assert(U3: Rabs (Z2R b * IDR x1 +IDR x2 - (Z2R b* IDR y1+ IDR y2 + r0 )) < 1).
  apply make_k_digit_tech2 with (1:=H9) (2:=H3).
  apply Rmult_lt_reg_l with (Z2R b); [ZR0compat|].
  apply Rlt_trans with (Z2R b - 1); [| R_to_Z].
  apply Rle_lt_trans with (2:=Hxy); right.
  replace (Z2R b * IDR x1 + IDR x2 + vx - (Z2R b * IDR y1 + IDR y2 + r0)) 
    with (Z2R b * ((IDR x2 + vx) / Z2R b - (- 1 + (IDR y2 + r0) / Z2R b)))
    by (rewrite R1; field_IDR).
  rewrite Rabs_mult; rewrite (Rabs_right (Z2R b)); [simpl; ring|ZR0compat].

replace (Z2R b * IDR x1 + IDR x2 - (Z2R b * IDR y1 + IDR y2 + r0))
  with  (IDR x2 - (- Z2R b  + IDR y2 + r0)) in U3
  by (rewrite R1; simpl_IDR; ring). 
rabs_elim.
apply lt_Z2R; simpl_Z2R; fold_IDR.
apply Rlt_le_trans with (1 - IDR x2 - Z2R b + r0); [dec_tac; fourier|].
assert (- Z2R b + 2 <= IDR x2); [| dec_tac; fourier].
IDR_to_Z.
generalize (Hd2 O); unfold Str_nth; simpl.
intros U4; apply U4; omega.

replace ((IDR (y2 - Dbase)%D + r0) / Z2R b) with 
  (-1 + (IDR y2 + r0) / Z2R b) by field_IDR; assumption.
Qed.

Theorem md_tail_bounds_right : 
  forall x vx vtl, 
    Srepresents x vx ->
    Srepresents (tl (make_digit x)) vtl ->
    (IDZ (hd (make_digit x)) < b - 1)%Z ->
    2*Z2R b * vtl <= Z2R b + 2.  (* b+1 ?*)
Proof.
assert (Hbpos:= base_positive).
assert (HZ2Rbpos : 0 <= Z2R b) by IDR_to_Z. 
intros x vx vtl Hrepx.
destruct x as [x1 [x2 x'']].
simpl.
unfold Dle_bool.
compare_gt_tac G1.
  assert (U1 : 2*IDR x2 <= - Z2R b-1) by IDR_to_Z.
  compare_gt_tac G2; simpl; simpl_IDZ; intros Htl Hhd;
    inversion Htl; subst.
    replace (2 * Z2R b * ((IDR (x2 + Dbase)%D + r) / Z2R b)) with 
      (2 * IDR x2 + 2* Z2R b + 2*r) by field_IDR.
    dec_tac; fourier.
    
    replace (2 * Z2R b * ((IDR x2  + r) / Z2R b)) with 
      (2 * IDR x2 +  2*r) by field_IDR.
    dec_tac; fourier.

  compare_gt_tac G2.
    compare_gt_tac G3; simpl; simpl_IDZ; intros Htl Hhd.
      inversion Htl; subst.
      replace (2 * Z2R b * ((IDR (x2 - Dbase)%D + r) / Z2R b)) with 
	(2 * IDR x2 - 2* Z2R b + 2*r) by field_IDR.
      assert (IDR x2 <= Z2R b - 1).
	inversion Hrepx.
	inversion H3; IDR_to_Z.
      dec_tac; fourier.

      absurd (IDZ x1 < b- 1)%Z; omega.

    simpl; intros Htl Hhd.
    inversion Htl; subst.

    replace (2 * Z2R b * ((IDR x2  + r) / Z2R b)) with 
      (2 * IDR x2 +  2*r) by field_IDR.
    assert (2*IDR x2 <= Z2R b) by IDR_to_Z.
    dec_tac; fourier.
Qed.

(*On devrait ne plus avoir de probleme en considérant des séries ou la somme partielle es toujours inférieure à 
  (b^2-3b-2)/(2*b^2) en valeur absolue
  le reste serait alors toujours plus petit que (b+2)/(2*b) en valeur absolue
*)
Theorem md_tech1 : 
  forall x r vx vr, 
    Srepresents x vx -> 
    Srepresents r vr -> 
    -(Z2R b *Z2R b-3*Z2R b -2) <= 2*Z2R b * Z2R b * vx <= Z2R b *Z2R b-3*Z2R b -2 ->
    -(Z2R b +2) <= 2*Z2R b * vr  <= Z2R b +2 ->
    - (2*Z2R b*Z2R b -Z2R b-2) <= 2*Z2R b*Z2R b * (vr+vx) <= 2*Z2R b*Z2R b -Z2R b-2.
Proof.   
assert (Hbpos := base_positive).
intros x r vx vr Hrepx Hrepr Hvx Hvr.

assert (U1 : 0 < Z2R b) by IDR_to_Z.

assert (U2 :  - (Z2R b *Z2R b + 2*Z2R b) <= 
  2 * Z2R b *Z2R b* vr <= Z2R b * Z2R b + 2*Z2R b).
 replace ( - (Z2R b * Z2R b + 2 * Z2R b)) with 
   (Z2R b * (- (Z2R b + 2))) by ring.
 replace (Z2R b * Z2R b + 2 * Z2R b) with 
   (Z2R b * (Z2R b + 2)) by ring.
 replace (2*Z2R b * Z2R b * vr) with 
   (Z2R b * (2*Z2R b *vr)) by ring.
split; apply Rmult_le_compat_l; intuition.

 match goal with 
   id1 : - ?r1 <= ?r0 * vx <= ?r1,
   id2 : - ?r2 <= ?r0 * vr <= ?r2
   |- -?r3 <= ?r0*(vr+vx) <= ?r3 =>
     rewrite (Rmult_plus_distr_l r0); 
     replace r3 with (r1+r2); [ dec_tac; split; fourier| ring]
  end.
Qed.

Theorem md_tech2 : 
  forall x r vx vr, 
    Srepresents x vx -> 
    Srepresents r vr -> 
    -(Z2R b *Z2R b-3*Z2R b -2) <= 2*Z2R b * Z2R b * vx <= Z2R b *Z2R b-3*Z2R b -2 ->
    -(Z2R b +2) <= 2*Z2R b * vr  <= Z2R b +2 ->
    Srepresents (Sadd r x) (vr+vx).
Proof.   
pose (Hbpos := base_positive).
intros x r vx vr Hrepx Hrepr Hvx Hvr.
apply Sadd_correct; assumption || omega || idtac.
generalize (md_tech1 x r vx vr Hrepx Hrepr Hvx Hvr).
intros [U1 U2].
assert (U3: -(2 * Z2R b * Z2R b) <= 2 * Z2R b * Z2R b * (vr+vx) <= 2 * Z2R b * Z2R b).
 assert (U4 : 2 * Z2R b * Z2R b - Z2R b - 2 <= 2 * Z2R b * Z2R b) by IDR_to_Z.
 split;[ apply Rle_trans with (2:= U1)|
   apply Rle_trans with (1:= U2)]; intuition.
assert (U4 : 0 < Z2R b) by IDR_to_Z.
assert (U5 : 0 < 2 * Z2R b * Z2R b) by ZR0compat.
dec_tac; split; (apply Rmult_le_reg_l with ( 2 * Z2R b * Z2R b); [assumption|]).
replace (2 * Z2R b * Z2R b * -1) with (- ( 2 * Z2R b * Z2R b)) by ring; assumption.
rewrite Rmult_1_r; assumption.
Qed.

Theorem make_digit_remainder : 
  forall x vx r vr k r', 
    Srepresents x vx -> 
    Srepresents r vr -> 
    -(Z2R b *Z2R b-3*Z2R b -2) <= 2*Z2R b * Z2R b * vx <= Z2R b *Z2R b-3*Z2R b -2 ->
    -(Z2R b +2) <= 2*Z2R b * vr  <= Z2R b +2 ->
    (k::r') = make_digit (Sadd r x) ->
    - (Z2R b+2)<= 2*Z2R b * ((vr+vx) * Z2R b -IDR k) <= Z2R b +2.
Proof.
intros x vx r vr k r' Hrepx Hrepr Hvx Hvr.
assert (Hrepadd : Srepresents (Sadd r x) (vr+vx)).
  apply md_tech2; assumption.
generalize_all (Sadd r x) x'.
assert (Hvx': -(2*Z2R b*Z2R b -Z2R b-2) <= 2*Z2R b*Z2R b * (vr+vx) <= 2*Z2R b*Z2R b -Z2R b-2).
  apply (md_tech1 x r); assumption.
generalize_all (vr+vx) vx'.
clear Hvr Hvx Hrepx Hrepr x vx r vr.

destruct x' as [k1 [k2 x]].
unfold make_digit, Dle_bool.
compare_gt_tac G1; compare_gt_tac G2; try (compare_gt_tac G3);
  intros U0; injection U0; clear U0; intros;  
    inversion_to_variable x; subst.

assert (U1 : -Z2R b <= 2 * IDR k2+2*Z2R b) by IDR_to_Z.
assert (U2 : 2 * IDR k2+2*Z2R b<= Z2R b) by IDR_to_Z.
match goal with 
  |-  _ <= ?r <= _ =>
    replace r with (2* IDR k2 +2*r0 + 2*Z2R b) by field_IDR;
       dec_tac; split; fourier
end.

absurd(- (2 * Z2R b * Z2R b - Z2R b - 2) <= 
  2 * Z2R b * Z2R b * ((IDR k1 + (IDR k2 + r0) / Z2R b) / Z2R b)); [|intuition].
 apply RIneq.Rlt_not_le.
 replace (IDR k1) with (-Z2R b+1) by IDR_to_Z.
 replace (2 * Z2R b * Z2R b * ((-Z2R b+1 + (IDR k2 + r0) / Z2R b) / Z2R b))
   with (-2 * Z2R b * Z2R b+2*Z2R b + 2*IDR k2 + 2*r0) by field_IDR.
 replace (- (2 * Z2R b * Z2R b - Z2R b - 2)) with 
   (- 2 * Z2R b * Z2R b + Z2R b + 2) by ring.
 assert (U1 : 2*IDR k2 < - Z2R b) by IDR_to_Z.
 dec_tac; fourier.

assert (U1 : -Z2R b <= 2 * IDR k2-2*Z2R b) by IDR_to_Z.
assert (U2 : 2 * IDR k2 -2*Z2R b<= Z2R b) by IDR_to_Z.
match goal with 
  |-  _ <= ?r <= _ =>
    replace r with (2* IDR k2 +2*r0-2*Z2R b) by field_IDR;
       dec_tac; split; fourier
end.

absurd(2 * Z2R b * Z2R b * ((IDR k1 + (IDR k2 + r0) / Z2R b) / Z2R b) <=
         2 * Z2R b * Z2R b - Z2R b - 2); [|intuition].
 apply RIneq.Rlt_not_le.
 replace (IDR k1) with (Z2R b-1) by IDR_to_Z.
 replace (2 * Z2R b * Z2R b * ((Z2R b-1 + (IDR k2 + r0) / Z2R b) / Z2R b))
   with (2 * Z2R b * Z2R b-2*Z2R b + 2*IDR k2 + 2*r0) by field_IDR.
 assert (U1 : Z2R b < 2*IDR k2) by IDR_to_Z.
 dec_tac; fourier.

assert (U1 : -Z2R b <= 2 * IDR k2) by IDR_to_Z.
assert (U2 : 2 * IDR k2 <= Z2R b) by IDR_to_Z.
match goal with 
  |-  _ <= ?r <= _ =>
    replace r with (2* IDR k2 +2*r0) by field_IDR;
      dec_tac; split; fourier
end.
Qed.


Theorem make_digit_bounds  :
  forall x vx vx',
    Srepresents x vx -> 
    - 1  < vx < 1 ->
    let (x1  , x') := make_digit x in 
      Srepresents x' vx' ->
      vx = (vx' + IDR x1)/Z2R b ->
      - 1 < vx' < 1.
Proof.
assert (Hbpos := base_positive).
intros x vx vx' Hrepx Hbounds.
unfold make_digit, Dle_bool.
destruct x as [x1 [x2 x']].
compare_gt_tac G1; compare_gt_tac G2; try compare_gt_tac G3.

 (*case 2*x2 < -b /\  -b+1 < x1*)
  simpl_IDR; fold_IDR.
  intros Hrepx' Hvx.
  inversion Hrepx'; subst.
  simpl_IDR; fold_IDR.
  split.
    apply Rlt_div_l; [R_to_Z |].
    apply Rlt_le_trans with 0; [R_to_Z|].
    assert (- Z2R b +1  <= IDR x2); [| dec_tac; fourier]. 
    inversion Hrepx; inversion H5; IDR_to_Z.
  
    apply Rlt_div_r; [R_to_Z |].
    assert (IDR x2 < -1) by IDR_to_Z; dec_tac; fourier.

 (*case 2*x2 < -b /\  x1  = -b+1*)
  replace (IDR x1) with (- Z2R b + 1) ; [| inversion Hrepx; IDR_to_Z].
  intros Hrepx' Hvx.
  split.
    replace vx'  with (vx * Z2R b + Z2R b - 1) by (rewrite Hvx; field_Z).
    assert (-1 * Z2R b < vx * Z2R b); [| fourier].
    apply Rmult_lt_compat_r; [R_to_Z | intuition].

    inversion Hrepx'; subst.
    apply Rlt_div_r; [R_to_Z |].
    assert (IDR x2 < -1) by IDR_to_Z.
    assert (4 < Z2R b) by IDR_to_Z; dec_tac; fourier.
    
(* case  b < 2*x2 /\ x1 < b-1 *)
  simpl_IDR; fold_IDR.
  intros Hrepx' Hvx.
  inversion Hrepx'; subst.
  simpl_IDR; fold_IDR.
  split.
    apply Rlt_div_l; [R_to_Z |].
    assert (1 < IDR x2) by IDR_to_Z; dec_tac; fourier.

    apply Rlt_div_r; [R_to_Z |].
    apply Rle_lt_trans with 0; [|R_to_Z].
    assert (IDR x2 <= Z2R b - 1); [| dec_tac; fourier].
    inversion Hrepx; inversion H5; IDR_to_Z.
  
(* case  b < 2*x2 /\ x1 = b-1 *)
  replace (IDR x1) with (Z2R b - 1) ; [| inversion Hrepx; IDR_to_Z].
  intros Hrepx' Hvx.
  split.
    inversion Hrepx'; subst.
    apply Rlt_div_l; [R_to_Z |].
    assert (1 < IDR x2 ) by IDR_to_Z.
    assert (4 < Z2R b) by IDR_to_Z; dec_tac; fourier.
  
    replace vx'  with (vx * Z2R b + - Z2R b + 1) by (rewrite Hvx; field_Z).
    assert (vx * Z2R b < 1 * Z2R b); [| fourier].
    apply Rmult_lt_compat_r; [R_to_Z | intuition].

(* case - b <= 2*x2 <= b *)
  inversion 1; intros U1; subst.
  split.
    apply Rlt_div_l; [R_to_Z |].
    assert ( -Z2R b + 1 < IDR x2 ) by IDR_to_Z.
    dec_tac; fourier.

    apply Rlt_div_r; [R_to_Z |].
    assert ( IDR x2 < Z2R b - 1) by IDR_to_Z.
    dec_tac; fourier.
Qed.

Theorem make_digit_add :
  forall x k x' vx vy,
    make_digit x = (k::x') ->
    Srepresents x vx ->
    (-1 <= vx + vy <= 1 -> 
      -(Z2R b-2) <= vy*(2*Z2R b*Z2R b) <= Z2R b-2 -> 
      -1 <= (vx + vy)*Z2R b - IDR k <= 1)%R.
Proof.
pose (Hbpos := base_positive).
intros x k x' vx vy Hmd_eq Hrep Hsum Hvy.
assert (U1 : (0 < Z2R b)%R).
IDR_to_Z.
assert (U2: (0< 2*Z2R b)%R).
fourier.

destruct x as [k1 [k2 x]].
inversion_to_variable x; subst.

assert (U3 : ((-1*Z2R b +- IDR k1) *(2 * Z2R b) <= 
  IDR k2 + IDR k2 + 2 * r0 + vy * (2 * Z2R b * Z2R b) <= 
  (1*Z2R b + -IDR k1) *(2 * Z2R b))%R).
replace (IDR k2 + IDR k2 + 2 * r0 + vy * (2 * Z2R b * Z2R b))%R with 
  ((((IDR k1 + (IDR k2 + r0) / Z2R b) / Z2R b + vy)*Z2R b+ -IDR k1)*(2*Z2R b))%R;
  [idtac| field_IDR].
dec_tac; split; 
  (apply Rmult_le_compat_r; [fourier|
    apply Rplus_le_compat_r; 
  apply Rmult_le_compat_r; assumption || fourier]).

replace (((IDR k1 + (IDR k2 + r0) / Z2R b) / Z2R b + vy) * Z2R b - IDR k)%R with 
  (((IDR k1-IDR k)*(2*Z2R b) + (IDR (k2+k2)%D + 2*r0) + vy*(2*Z2R b*Z2R b))/
    (2*Z2R b))%R; [idtac| simpl_IDR; field_IDR].
apply Rle_div_1.
IDR_to_Z.
generalize Hmd_eq; clear Hmd_eq.
unfold make_digit, Dle_bool.

compare_gt_tac G1; compare_gt_tac G2; try compare_gt_tac G3;
intros U5; injection U5; clear U5; intros U5 U6; subst; simpl_IDR;
match goal with 
  |- (_ <= (?v0 - ?v0) * ?v1 + ?v2 +?v3 <= _ )%R =>
    replace ((v0-v0)*v1+v2+v3)%R  with (v2+v3)%R;[idtac|ring]
  | |- (_ <= (?v0 - (?v0+1)) * ?v1 + ?v2 +?v3 <= _ )%R =>
    replace ((v0-(v0+1))*v1+v2+v3)%R  with (-v1+(v2+v3))%R;[idtac|ring]
  | |- (_ <= (?v0 - (?v0-1)) * ?v1 + ?v2 +?v3 <= _ )%R =>
    replace ((v0-(v0-1))*v1+v2+v3)%R  with (v1+(v2+v3))%R;[idtac|ring]
  | _ => idtac
end.

split.
assert (U4: (-2*Z2R b +2 <= IDR k2 + IDR k2)%R).
IDR_to_Z.
dec_tac.
apply Rle_trans with ((2 * Z2R b) + (-(2*Z2R b) + -(2*Z2R b)))%R.
fourier.
apply Rplus_le_compat_l.
apply Rplus_le_compat; fold_IDR; fourier.

match goal with 
  |- (_ <= ?a )%R => 
    pattern a at 3; replace a with (a+((-Z2R b+2)+(Z2R b-2)))%R; [idtac| ring]  
end.
apply Rplus_le_compat_l.
assert (U4: (IDR k2 + IDR k2<= -Z2R b)%R) by IDR_to_Z.
repeat apply Rplus_le_compat; intuition; fourier.

assert (U4: (IDR k= (-Z2R b +1))%R).
IDR_to_Z.
dec_tac; split.
replace (-(2 * Z2R b))%R with ((-1 * Z2R b + -IDR k) * (2 * Z2R b))%R;
  [ assumption| rewrite U4; ring].
assert (U5: (IDR k2 + IDR k2 <= Z2R b)%R);
  [IDR_to_Z |  fold_IDR; fourier].

split.
match goal with 
  |- (?a <= _ )%R => 
    pattern a at 1; replace a with (a+((Z2R b+-2)+(-(Z2R b-2))))%R; [idtac| ring]  
end.
apply Rplus_le_compat_l.
assert (U4: (Z2R b <= (IDR k2 + IDR k2))%R);
  [IDR_to_Z| idtac].
repeat apply Rplus_le_compat; intuition; fourier.
assert (U4: (IDR k2 + IDR k2<= 2*Z2R b -2)%R).
IDR_to_Z.
dec_tac.
apply Rle_trans with (- (2 * Z2R b) + (2*Z2R b + 2*Z2R b))%R.
apply Rplus_le_compat_l.
apply Rplus_le_compat; fold_IDR; fourier.
fourier.

assert (U4: (IDR k= (Z2R b -1))%R) by IDR_to_Z.
dec_tac; split.
assert (U5: (-Z2R b <= IDR k2 + IDR k2)%R) by IDR_to_Z.
fold_IDR; fourier.
pattern (2 * Z2R b)%R at 2; 
  replace (2 * Z2R b) %R with ((1 * Z2R b + - IDR k) * (2 * Z2R b))%R;
  [ assumption| rewrite U4; ring].

assert (U4: ( - Z2R b <= IDR k2 + IDR k2 <= Z2R b)%R).
split; IDR_to_Z.
fold_IDR; dec_tac; split; fourier.
Qed.

Theorem series_bound : 
  forall b,
    (0 < b)%Z ->
    (Z2R b-2 <= 2* Z2R b*Z2R b)%R.
Proof.  
intros b' Hbpos.
IDR_to_Z.
apply Zle_trans with (1* b')%Z.
omega.
apply Zmult_le_compat_r; omega.
Qed.

Ltac case_md d r' Hmd :=
  match goal with
    |- context [ make_digit ?r ] =>
      caseEq (make_digit r); intros d r' Hmd
  end.

Ltac md_eq_hyp Hmd_eq:= 
  match goal with 
    | id1 : make_digit ?r = (?k::?r'), id2 : Srepresents ?r ?vr
      |- Srepresents (?k::_) _ =>
	assert (Hmd_eq : Srepresents (k::r') vr); 
	  [rewrite <- id1; apply make_digit_eq; assumption  | idtac]
  end.    

Ltac md_add_hyp Hmd_add:= 
  match goal with 
    | id : make_digit ?r = (?k::?r') |- Srepresents (?k::_) (?vx+?vy)%R =>
      assert (Hmd_add : (-1 <= (vx+vy)*Z2R b - IDR k <= 1)%R); 
	[apply make_digit_add with (2:=id); (assumption || omega|| idtac) | idtac]
  end.    

Ltac make_digit_tac d r' U1 U2 U3 U4 series_correct := 
  match goal with
    |- Srepresents (match  make_digit ?r with k::r' => _ end) (?v1+?v2)%R => 
      caseEq (make_digit r); intros d r' U1;
	assert (U2 : Srepresents r v1); [clear series_correct | 
	  assert (U3: Srepresents (d::r') v1) ;
	    [ clear series_correct; rewrite <- U1; apply make_digit_eq; (assumption || omega|| idtac)|
	      assert (U4: (-1<= (v1+v2)*Z2R b - IDR d<=1)%R); 
		[ clear series_correct ; apply make_digit_add with (1:=U1); 
		  (assumption|| omega ||idtac) | idtac]]]
end.

Theorem md_ultimate : 
  forall x vx r vr k r', 
    Srepresents x vx -> 
    Srepresents r vr -> 
    -(Z2R b *Z2R b-3*Z2R b -2) <= 2*Z2R b * Z2R b * vx <= Z2R b *Z2R b-3*Z2R b -2 ->
    -(Z2R b +2) <= 2*Z2R b * vr  <= Z2R b +2 ->
    make_digit (Sadd r x) = (k::r') ->
    - (Z2R b+2)<= 2*Z2R b * ((vr+vx) * Z2R b -IDR k) <= Z2R b +2 /\
    (- b < IDZ k < b)%Z /\
    Srepresents r' ((vr+vx)*Z2R b-IDR k).
Proof.
intros x vx r vr k r' Hrepx Hrepr Hvx Hvr.
assert (Hrepadd : Srepresents (Sadd r x) (vr+vx)).
  apply md_tech2; assumption.
generalize_all (Sadd r x) x'.
assert (Hvx': -(2*Z2R b*Z2R b -Z2R b-2) <= 2*Z2R b*Z2R b * (vr+vx) <= 2*Z2R b*Z2R b -Z2R b-2).
  apply (md_tech1 x r); assumption.
generalize_all (vr+vx) vx'.
clear Hvr Hvx Hrepx Hrepr x vx r vr.

destruct x' as [k1 [k2 x]].
unfold make_digit, Dle_bool;
  compare_gt_tac G1; compare_gt_tac G2; try compare_gt_tac G3;
  intros U0; injection U0; clear U0; intros;  
    inversion_to_variable x; subst.

split. 
  assert (U1 : -Z2R b <= 2 * IDR k2+2*Z2R b) by IDR_to_Z.
  assert (U2 : 2 * IDR k2+2*Z2R b<= Z2R b) by IDR_to_Z.
  match goal with 
    |-  _ <= ?r <= _ =>
      replace r with (2* IDR k2 +2*r0 + 2*Z2R b) by field_IDR;
	dec_tac; split; fourier
  end.
  
 split; [simpl_IDZ; omega| ].
 replace ((IDR k1 + (IDR k2 + r0) / Z2R b) / Z2R b * Z2R b - IDR (k1 - 1)%D) with
   ((IDR (k2 + Dbase)%D + r0)/ Z2R b) by field_IDR.
 constructor; assumption || simpl_IDZ; omega.

absurd(- (2 * Z2R b * Z2R b - Z2R b - 2) <= 
  2 * Z2R b * Z2R b * ((IDR k + (IDR k2 + r0) / Z2R b) / Z2R b)); [|intuition].
 apply RIneq.Rlt_not_le.
 replace (IDR k) with (-Z2R b+1) by IDR_to_Z.
 replace (2 * Z2R b * Z2R b * ((-Z2R b+1 + (IDR k2 + r0) / Z2R b) / Z2R b))
   with (-2 * Z2R b * Z2R b+2*Z2R b + 2*IDR k2 + 2*r0) by field_IDR.
 replace (- (2 * Z2R b * Z2R b - Z2R b - 2)) with 
   (- 2 * Z2R b * Z2R b + Z2R b + 2) by ring.
 assert (U1 : 2*IDR k2 < - Z2R b) by IDR_to_Z.
 dec_tac; fourier.

split.
  assert (U1 : -Z2R b <= 2 * IDR k2-2*Z2R b) by IDR_to_Z.
  assert (U2 : 2 * IDR k2 -2*Z2R b<= Z2R b) by IDR_to_Z.
  match goal with 
    |-  _ <= ?r <= _ =>
      replace r with (2* IDR k2 +2*r0-2*Z2R b) by field_IDR;
	dec_tac; split; fourier
  end.
 split; [simpl_IDZ; omega| ].
 replace ((IDR k1 + (IDR k2 + r0) / Z2R b) / Z2R b * Z2R b - IDR (k1 + 1)%D) with
   ((IDR (k2 - Dbase)%D + r0)/ Z2R b) by field_IDR.
 constructor; assumption || simpl_IDZ; omega.

absurd(2 * Z2R b * Z2R b * ((IDR k + (IDR k2 + r0) / Z2R b) / Z2R b) <=
         2 * Z2R b * Z2R b - Z2R b - 2); [|intuition].
 apply RIneq.Rlt_not_le.
 replace (IDR k) with (Z2R b-1) by IDR_to_Z.
 replace (2 * Z2R b * Z2R b * ((Z2R b-1 + (IDR k2 + r0) / Z2R b) / Z2R b))
   with (2 * Z2R b * Z2R b-2*Z2R b + 2*IDR k2 + 2*r0) by field_IDR.
 assert (U1 : Z2R b < 2*IDR k2) by IDR_to_Z.
 dec_tac; fourier.

split.
  assert (U1 : -Z2R b <= 2 * IDR k2) by IDR_to_Z.
  assert (U2 : 2 * IDR k2 <= Z2R b) by IDR_to_Z.
  match goal with 
    |-  _ <= ?r <= _ =>
      replace r with (2* IDR k2 +2*r0) by field_IDR;
	dec_tac; split; fourier
  end.
  
  intuition.
  replace ((IDR k + (IDR k2 + r0) / Z2R b) / Z2R b * Z2R b - IDR k) with 
    ( (IDR k2 + r0) / Z2R b) by field_IDR; assumption.
Qed.

Ltac md_tac2 d r' vr' U1 U2 series_correct :=
  match goal with 
    |- Srepresents (series_body _ (Sadd ?r1 ?x1)) (?vr+?vx+?vs)%R => 
      unfold series_body; 
	caseEq (make_digit (Sadd r1 x1)); intros d r' U1;
	  assert(U2 :
	    - (Z2R b+2)<= 2*Z2R b * ((vr+vx) * Z2R b -IDR d) <= Z2R b +2 /\
	    (- b < IDZ d < b)%Z /\
	    Srepresents r' ((vr+vx)*Z2R b-IDR d)) ;
	  [clear series_correct; apply md_ultimate with (x:=x1) (r:=r1); 
	    assumption || (simpl_IDZ; omega) || idtac |
	  replace (vr+vx+vs) with ((IDR d+ ((vr+vx)*Z2R b-IDR d+ vs*Z2R b))/ Z2R b);
	    [ | clear series_correct; try field_IDR ]]
  end.





Theorem splitted_series_bound : 
  forall x y,
    - (Z2R b-4) <= 2* Z2R b * (x+y) <= Z2R b-4 ->
    - (Z2R b-2) <= 2* Z2R b * Z2R b* y <= Z2R b-2 ->
    - ( Z2R b* Z2R b -3* Z2R b -2) <= 2* Z2R b * Z2R b* x <= Z2R b* Z2R b -3* Z2R b -2.
Proof.
assert (Hbpos := base_positive).
intros x y Hxy Hy .

rabs_intros.

rewrite Rabs_mult.

apply Rle_trans with ( Rabs (2 * Z2R b * Z2R b) * (Rabs ( x + y) + Rabs( y))).
  apply Rmult_le_compat_l; ZR0compat; IDR_to_Z.
  generalize (Rabs_triang_inv x (-y)).
  rewrite Rabs_Ropp.
  replace (x - - y) with (x+y) by ring.
  intros; fourier.

  apply Rle_trans with (Z2R b * (Z2R b - 4) + (Z2R b - 2)).
    rewrite Rmult_plus_distr_l.
    apply Rplus_le_compat.
    replace ( Rabs (2 * Z2R b * Z2R b) * Rabs (x + y)) with 
      (Z2R b * Rabs (2 * Z2R b * (x + y))).
    apply  Rmult_le_compat_l; ZR0compat.
    repeat rewrite Rabs_mult.
    rewrite (Rabs_right (Z2R b)); ZR0compat.
    rewrite <- Rabs_mult; assumption.
    
    right; ring.
Qed.

Theorem splitted_series_bound_rabs : 
  forall x y,
    2* Z2R b * Rabs (x+y) <= Z2R b-4 ->
    2* Z2R b * Z2R b* Rabs y <= Z2R b-2 ->
    2* Z2R b * Z2R b*Rabs x <= Z2R b* Z2R b -3* Z2R b -2.
Proof.
assert (Hbpos := base_positive).
intros x y Hxy Hy .

apply Rle_trans with ( 2 * Z2R b * Z2R b * ( Rabs (x + y) + Rabs y)).
  apply Rmult_le_compat_l; ZR0compat; IDR_to_Z.
  generalize (Rabs_triang_inv x (-y)).
  rewrite Rabs_Ropp.
  replace (x - - y) with (x+y) by ring.
  intros; fourier.

  apply Rle_trans with (Z2R b * (Z2R b - 4) + (Z2R b - 2)).
    replace (2 * Z2R b * Z2R b * (Rabs (x + y) + Rabs y)) with 
      (Z2R b * (2 * Z2R b * Rabs (x + y)) + 2 * Z2R b * Z2R b * Rabs y) by ring.
    apply Rplus_le_compat.
    apply  Rmult_le_compat_l; ZR0compat.
    assumption.
    
    right; ring.
Qed.




End MAKE_DIGIT_CORRECT.
