Require Export make_digit_correct.

Module MULT_CORRECT(Nums : NUMS).

Module Make_digit_correct := MAKE_DIGIT_CORRECT(Nums).
Export Make_digit_correct.

Theorem mult_lemma : 
  forall k x,
    (-b < IDZ k < b)%Z ->
    Rabs (2 * Z2R b * Z2R b * (IDR k / Z2R b * x)) <=
    Rabs (x*(2 * Z2R b * Z2R b)).
Proof.
intros k x Hk.
assert (H : 0 < Z2R b) by IDR_to_Z.

replace (2 * Z2R b * Z2R b * (IDR k / Z2R b * x)) with 
    (x * (2 * Z2R b * IDR k)) by field_IDR.
elim (Rle_or_lt 0 x); intros G1.
  elim (Rle_or_lt 0 (IDR k)); intros G2.
    repeat rewrite Rabs_right; ZR0compat.
    repeat apply Rmult_le_compat_l; ZR0compat; IDR_to_Z.
    
    replace (x * (2 * Z2R b * IDR k)) with (- (x * (2 * Z2R b * - IDR k))) by ring.
    rewrite Rabs_left1; ZR0compat. 
    rewrite Rabs_right; ZR0compat.
    rewrite Ropp_involutive.
    repeat apply Rmult_le_compat_l; ZR0compat; IDR_to_Z.
    
  elim (Rle_or_lt 0 (IDR k)); intros G2.
    replace (x * (2 * Z2R b * IDR k)) with (- (-x * (2 * Z2R b * IDR k))) by ring.
    replace (x * (2 * Z2R b * Z2R b)) with (- (-x * (2 * Z2R b * Z2R b))) by ring.
    repeat rewrite Rabs_left1; ZR0compat. 
    repeat rewrite Ropp_involutive.
    repeat apply Rmult_le_compat_l; ZR0compat; IDR_to_Z.

    replace (x * (2 * Z2R b * IDR k)) with (-x * (2 * Z2R b * - IDR k)) by ring.
    replace (x * (2 * Z2R b * Z2R b)) with (- (-x * (2 * Z2R b * Z2R b))) by ring.
    rewrite Rabs_right; ZR0compat.
    rewrite Rabs_left1; ZR0compat. 
    rewrite Ropp_involutive.
    repeat apply Rmult_le_compat_l; ZR0compat; IDR_to_Z.
Qed.    

Theorem mult_series_correct : 
  forall x y r vx vy vr,
    Srepresents x vx ->
    Srepresents y vy ->
    Srepresents r vr ->
    - (Z2R b + 2) <= 2 * Z2R b * vr <= Z2R b + 2 ->
    (-(Z2R b-2) <= vy *(2*Z2R b*Z2R b) <= Z2R b-2 ->
      -1 <= vr + vx*vy <= 1 ->
      Srepresents (mult_series x y r) (vr + vx * vy))%R.
Proof.
pose (Hbpos := base_positive).
assert (H2bpos: (0<Z2R b)%R) by abstract IDR_to_Z.

cofix.
intros x y r vx vy vr Hrepx Hrepy Hrepr Hr Hvy Hsum.
str_unfold (mult_series x y r).
simpl.
fold_decomp.
destruct x as [k1 x']; inversion Hrepx; subst; simpl_IDZ.

replace (vr + (IDR k1 + r0) / Z2R b * vy) with 
  (vr + IDR k1 / Z2R b * vy + (r0 * vy / Z2R b)) by (unfold Rdiv; ring).

md_tac2 d r' vr' U1 U2 mult_series_correct. 

 apply mult_digit_correct; assumption || omega.

 rabs_intros.
 apply Rle_trans_trans with (2:=Hvy).
 apply mult_lemma; assumption.
 replace (Z2R b * Z2R b - 3 * Z2R b - 2) with (Z2R b * (Z2R b -3) -2) by ring.
 replace (Z2R b -2) with (Z2R b *1 -2) by ring.
 unfold Rminus; apply Rplus_le_compat_r.
 apply Rmult_le_compat_l; IDR_to_Z.

destruct U2 as [U2 [U3 U4]].
replace ((vr + IDR k1 / Z2R b * vy) * Z2R b - IDR d + r0 * vy / Z2R b * Z2R b) with 
  (((vr + IDR k1 / Z2R b * vy) * Z2R b - IDR d) + r0 * vy) by field_IDR.

cut (-1 <= (vr + IDR k1 / Z2R b * vy) * Z2R b - IDR d + r0 * vy <= 1).
intros U5.

constructor; assumption || apply mult_series_correct; assumption.

clear mult_series_correct.
rabs_intros.
replace 1 with ((Z2R b + 2)/(2*Z2R b) + (Z2R b - 2)/(2*Z2R b)) by field_IDR.
apply Rabs_add.
apply Rle_div_l; ZR0compat.
rewrite Rmult_comm.
rewrite <- (Rabs_right (2*Z2R b)); ZR0compat.
rewrite <- Rabs_mult; assumption.

rewrite Rabs_mult.
rewrite <- (Rmult_1_l ((Z2R b - 2) / (2 * Z2R b))).
apply Rmult_le_compat; ZR0compat.

apply Rle_div_l; ZR0compat.
apply Rle_trans with (2:=Hvy).
do 2 rewrite Rabs_mult.
replace (Rabs vy * (2 * Z2R b)) with 
   (Rabs vy * (Rabs (2 * Z2R b) *1)).
repeat apply Rmult_le_compat_l; ZR0compat.
rewrite Rabs_right; ZR0compat; IDR_to_Z.
rewrite (Rabs_right (2*Z2R b)); ZR0compat.
Qed.

Theorem Smult_correct : 
  forall vx vy x y,
    Srepresents x vx ->
    Srepresents y vy ->
    Srepresents (Smult x y) (vx*vy)%R.
Proof. 
pose (Hbpos := base_positive).
assert (U1 : 4 <= Z2R b) by IDR_to_Z.
intros vx vy x y Hrepx Hrepy.
unfold Smult.

generalize (Srepresents_interval _ _ Hrepx).
generalize (Srepresents_interval _ _ Hrepy).
intros Hvy Hvx.

assert  (U2: (-1 <= vx*vy <= 1)%R).
apply Rdiv_le_mult; intuition.

assert (U3 : (-1 <= vx * vy / Z2R b <= 1)%R).
apply Rle_div_1; [IDR_to_Z| dec_tac; split; fourier].

assert (R1: ((vx * vy) = 
  Z2R b*(Z2R b*(0+vx*((IDR 0%D+(IDR 0%D+vy)/Z2R b)/Z2R b))))%R); 
[field_IDR | rewrite R1].

apply multb_correct; omega || idtac.

apply multb_correct; omega || idtac.

apply mult_series_correct; assumption || omega || idtac.
constructor.
constructor; assumption || simpl_IDZ; omega.

simpl_IDR; rewrite Rplus_0_l.
apply Rle_div_1; [fourier |].
dec_tac; split; fourier.

simpl_IDZ; omega.

const_rep; reflexivity.

split; fourier.

simpl_IDR; repeat rewrite Rplus_0_l.
replace (vy / Z2R b /Z2R b* (2 * Z2R b * Z2R b))%R with (2*vy)%R by field_IDR.
dec_tac; split; fourier.

simpl_IDR; repeat rewrite Rplus_0_l.
replace (vx * (vy / Z2R b / Z2R b))%R with 
  (((vx * vy) / Z2R b) / Z2R b)%R by field_IDR.
apply  Rle_div_1; [IDR_to_Z|].
apply Rle_le_trans with 1; assumption || IDR_to_Z.

simpl_IDR; repeat rewrite Rplus_0_l.
replace (Z2R b * (vx * (vy / Z2R b / Z2R b)))%R with 
  (vx * vy / Z2R b)%R; [assumption|field_IDR].
rewrite <- R1; assumption.
Qed.

Theorem mult_correct : 
  forall r1 r2 x y,
    represents x r1 ->
    represents y r2 ->
    represents (mult x y) (r1*r2)%R.
Proof.
intros r1 r2 x y.
unfold mult.
case y; case x; clear x y.
intros e1 x' e2 y' Hrepx' Hrepy'.
inversion Hrepx'; inversion Hrepy'; subst.
generalize (Srepresents_interval _ _ H2).
generalize (Srepresents_interval _ _ H6).
intros Hr0 Hr.
pose (Hbpos := base_positive).
assert (U1: 4 <= Z2R b) by IDR_to_Z.
assert (U2: (-1 <= r0 / Z2R b <= 1)%R).
apply Rle_div_1.
IDR_to_Z.
split; dec_tac; fourier.

replace (Z2R b ^ e1 * r * (Z2R b ^ e2 * r0))  with 
  (Z2R b ^(e1+e2+2)%nat* (0+r * ((IDR 0%D+(IDR 0%D+r0)/Z2R b)/ Z2R b))).
 constructor.
 apply mult_series_correct; assumption || omega || idtac.
 
  constructor; (simpl_IDZ; omega)|| idtac.
    constructor; assumption || (simpl_IDZ; omega).
    simpl_IDR; rewrite Rplus_0_l; assumption.
  
  const_rep; trivial.
  
  split; fourier.
  
  simpl_IDR; repeat rewrite Rplus_0_l.
  replace (r0 / Z2R b / Z2R b * (2 * Z2R b * Z2R b)) with (2*r0) by field_IDR.
  assert (2 <=  Z2R b - 2) by IDR_to_Z; split; dec_tac; fourier.

  simpl_IDR; repeat rewrite Rplus_0_l.
  apply Rdiv_le_mult; fourier || assumption || idtac.
  apply Rle_div_1.
    IDR_to_Z.
    generalize_all (r0/Z2R b) w; split; dec_tac; fourier.
  
  simpl_IDR; repeat rewrite Rplus_0_l.
  repeat rewrite pow_add.
  simpl; field; ZR0compat.
Qed.

End MULT_CORRECT.
