Require Export add_correct.

Module MULT_DIGIT_CORRECT(Nums : NUMS).

Module Add_correct := ADD_CORRECT(Nums).
Export Add_correct.

Lemma even_or_odd : 
  forall z, 
    (z > 0-> 
    (exists z', z = 2*z') \/ (exists z',  z = 2*z' +1))%Z.
Proof.
intros z Hzpos.
assert (U1: (2>0)%Z) by omega.
generalize (Z_div_mod_eq z 2 U1). 
generalize (Z_mod_lt z 2 U1). 
generalize (z mod 2)%Z.
generalize (z / 2)%Z.
intros z' r.

elim (Z_eq_dec r 0); intros G1 G2 G3.
left; exists z'; omega.
right; exists z'; omega.
Qed.

Theorem Ddiv_full : 
  forall m n, 
    (0 < IDZ n -> 
    let (q, r) := Ddiv_eucl m n in 
      IDZ m = IDZ n * IDZ q + IDZ r /\ 
      0 <= IDZ r < IDZ n)%Z.
Proof.
intros m n Hn.
generalize (div_eucl_IDZ m n).
caseEq ( Ddiv_eucl m n).
intros q r _.
assert (U1 : (IDZ n > 0)%Z) by omega.
generalize (Z_div_mod (IDZ m) (IDZ n) U1).
caseEq (Zdiv_eucl (IDZ m) (IDZ n)).
intros x y _ U2 U3; injection U3; intros U4 U5; subst; assumption.
Qed.

Ltac Ddiv_tac q r H1 H2 := 
  match goal with 
    |- context [ Ddiv_eucl ?m ?n] =>
      let Hn := fresh "H" in 
        assert (Hn : (0 < IDZ n)%Z) ;[try (simpl_IDZ; omega)|
          generalize (Ddiv_full m n Hn); clear Hn;
            caseEq (Ddiv_eucl m n); simpl_IDZ;
            intros q r _ [H1 H2]]
  end. 

Theorem find_mult_remainder_correct : 
  forall k x1, 
    ( -(b+1) <= 2*IDZ k <= b+1 ->
      -b < IDZ x1 < b -> 
      let (q,r'):= find_mult_remainder k x1 in 
        IDZ q * b + IDZ r' = IDZ k * IDZ x1 /\ 
        -b <= 2*IDZ r' <= b /\      
        -b < 2*IDZ q < b)%Z.
Proof.
assert (Hbpos := base_positive).
intros k x1 Hk Hx1.
unfold find_mult_remainder.

Ddiv_tac q r Hdiveq Hmod.
assert (U1 : (b > 0)%Z) by omega.

assert (U4 : (-b*b+1-2*IDZ r <= 2*IDZ q*b <= b*b-1-2*IDZ r)%Z). 
  assert (-b*b+1 <= 2*IDZ q*b + 2*IDZ r<= b*b-1)%Z; [| omega]. 
  replace (2 * IDZ q * b + 2 * IDZ r)%Z with (2* IDZ k * IDZ x1)%Z
    by (rewrite <- Zmult_assoc; rewrite Hdiveq; ring).
  replace (- b * b + 1)%Z with (- ((b+1)*(b-1)))%Z by ring.
  replace (b * b - 1)%Z with ((b+1)*(b-1))%Z by ring.
  apply Zmult_le_compat_sym; omega.

assert (U5 : (- b < 2 * (IDZ q + 1))%Z).
  assert ( - (b+2) < 2*IDZ q)%Z; [| omega].
  apply Zmult_lt_reg_r with b; [omega|].
  replace  (- (b + 2) * b)%Z with (-b*b-2*b)%Z by ring.
  omega.

assert (U6 : (2 * IDZ q < b)%Z).
  apply Zmult_lt_reg_r with b; [omega|].
  apply Zle_lt_trans with  (b * b - 1 - 2 * IDZ r)%Z; omega.

compare_gt_tac G1; [| compare_gt_tac G2]; simpl_IDZ;
(split; [rewrite Hdiveq; ring| split; split; try omega]).

elim (even_or_odd b U1); intros [b' Hb']; rewrite Hb' in *.
  assert (IDZ q < b'-1)%Z; [| omega].
  apply Zmult_lt_reg_r with (2*b')%Z; [omega|].
  assert (IDZ q * (2 * b') <= (b' - 1) * (2 * b')-1)%Z; [|omega].
  replace ( (b' - 1) * (2 * b')-1)%Z with ((2*b'-1)*b'+-(b'+1))%Z by ring.
  replace (IDZ q * (2 * b'))%Z with  (IDZ k * IDZ x1+-IDZ r)%Z;
    [| rewrite Hdiveq; ring].
  apply Zplus_le_compat; try omega; rewrite Zmult_comm.
  assert (-((2 * b' - 1) * b')<= IDZ x1 * IDZ k <= (2 * b' - 1) * b')%Z; 
    [| omega].
  apply Zmult_le_compat_sym; omega.
  
  assert (IDZ q <  b')%Z; [| omega].
  apply Zmult_lt_reg_r with (2*b'+1)%Z; try omega.
  assert (IDZ q * (2 * b'+1) <= b' * (2 * b'+1)-1)%Z; [| omega].
  replace ( b' * (2 * b'+1)-1)%Z with ((2*b')*(b'+1)+-(b'+1))%Z by ring.
  replace (IDZ q * (2 * b'+1))%Z with  (IDZ k * IDZ x1+-IDZ r)%Z;
    [ | rewrite Hdiveq; ring].
  apply Zplus_le_compat; try omega; rewrite Zmult_comm.
  assert (-((2 * b') * (b'+1))<= IDZ x1 * IDZ k <= (2 * b') * (b'+1))%Z; 
    [| omega]; apply Zmult_le_compat_sym; omega.

assert (Hq : (-b = 2*IDZ q)%Z).
  apply Zle_antisym; [|intuition].
  assert (-b-1 < 2*IDZ q)%Z;[ | omega]. 
  apply Zmult_lt_reg_r with (b); [omega | ].
  replace ((- b - 1) * b)%Z with (-b*b+-b)%Z by ring.
  apply Zlt_le_trans with (- b * b + 1 - 2 * IDZ r)%Z; omega.

assert (Hr'eq: (b = 2*IDZ r)%Z).
  apply Zle_antisym; try omega.
  replace (2*IDZ r)%Z with (2*(IDZ k*IDZ x1)+b*b)%Z.
  assert (b - b * b <= 2 * (IDZ k * IDZ x1))%Z; [| omega];
    replace (b - b * b)%Z with (-(b * (b-1)))%Z by ring;
      assert (-(b * (b-1)) <= 2 * (IDZ k * IDZ x1) <= b * (b-1))%Z;
        [| omega]; rewrite Zmult_assoc; apply Zmult_le_compat_sym; omega. 
  rewrite Hdiveq; rewrite Zmult_plus_distr_r.
  replace (2 * (b*IDZ q))%Z with (-b*b)%Z by (rewrite Hq; ring).
  ring.

replace (2 * (IDZ r - b))%Z with (-b)%Z by (rewrite Hr'eq; ring).
omega.
Qed.

Theorem mult_div_base_tech : 
  forall m n p ,
    (- p + 1 <= 2* m <= p + 1 -> 
    -p <= 2* n <= p-> 
    -p <= m + n <= p)%Z.
Proof.
intros m n p Hm Hn.
omega.
Qed.

Theorem mult_div_base_correct : 
  forall k r x vx,
    (-(b+1) <= 2*IDZ k <= b+1)%Z ->
    ( -b <= 2*IDZ r <= b)%Z ->
    Srepresents x vx -> 
    Srepresents (mult_div_base k r x) ((IDR k*vx+IDR r)/Z2R b)%R.
Proof.
assert (Hbpos := base_positive).
cofix.
intros k r x vx Hk Hr Hrepx.
str_unfold (mult_div_base k r x).
destruct x as [x1 x']; inversion Hrepx; subst.
simpl.
fold_decomp.
generalize (find_mult_remainder_correct _ _ Hk H4).
caseEq (find_mult_remainder k x1); intros q r' Hqr' [U1 [U2 U3]].
replace ((IDR k * ((IDR x1 + r0) / Z2R b) + IDR r) / Z2R b) with 
  ((IDR (q+ r)%D + (IDR k* r0+ IDR r')/ Z2R b)/ Z2R b).

constructor;[apply mult_div_base_correct| | ]; 
  clear mult_div_base_correct; try assumption.
apply Rle_div_1; [R_to_Z|].
assert (U4: - Z2R b - 1 <= 2 * IDR k <= Z2R b + 1) by (split; IDR_to_Z).
assert (U5 : - Z2R b <= 2 * IDR r <= Z2R b) by (split; IDR_to_Z).
split; elim (Z_lt_le_dec (IDZ k) 0); intros G1.
  apply Rle_trans with (IDR k + IDR r'); [IDR_to_Z| apply Rplus_le_compat_r].
  pattern (IDR k) at 1; replace (IDR k) with (-IDR k*-1) by ring.
  replace (IDR k*r0) with (-IDR k*-r0) by ring.
  apply Rmult_le_compat_l;[IDR_to_Z|dec_tac; fourier].

  apply Rle_trans with (- IDR k + IDR r'); [IDR_to_Z| apply Rplus_le_compat_r].
  replace (- IDR k) with (IDR k*-1) by ring.
  apply Rmult_le_compat_l;[IDR_to_Z|dec_tac; fourier].

  apply Rle_trans with (- IDR k + IDR r'); [apply Rplus_le_compat_r| IDR_to_Z].
  replace (-IDR k) with (-IDR k*1) by ring.
  replace (IDR k*r0) with (-IDR k*-r0) by ring.
  apply Rmult_le_compat_l;[IDR_to_Z|dec_tac; fourier].
  
  apply Rle_trans with (IDR k + IDR r'); [apply Rplus_le_compat_r | IDR_to_Z].
  pattern (IDR k) at 2; replace (IDR k) with (IDR k*1) by ring.
  apply Rmult_le_compat_l;[IDR_to_Z|dec_tac; fourier].

simpl_IDZ; omega.

replace (IDR k * ((IDR x1 + r0) / Z2R b)) with 
  ((IDR k * IDR x1 + IDR k * r0)/Z2R b) by (unfold Rdiv; ring).
simpl_IDR.
replace (Z2R (IDZ k) * Z2R (IDZ x1)) with (Z2R (IDZ k * IDZ x1)) by R_to_Z.
rewrite <- U1; field_IDR.
Qed.

Theorem mult_digit_correct : 
  forall k x vx,
    (-(b-1) <= IDZ k <= b-1)%Z ->
    Srepresents x vx -> 
    Srepresents (mult_digit k x) (IDR k/Z2R b*vx)%R.
Proof.
intros k x vx Hk Hrep.

assert (Hvx : (-1 <= vx <= 1)%R).
apply Srepresents_interval with (1:=Hrep).

unfold mult_digit.
unfold Dle_bool.

compare_gt_tac G1; [| compare_gt_tac G2]. 

assert (Heq: (IDR k / Z2R b *vx = (IDR (k-Dbase)%D * vx +IDR 0%D)/ Z2R b + vx)%R); 
  [field_IDR | rewrite Heq].
apply Sadd_correct; (assumption|| idtac).
apply mult_div_base_correct; (assumption||simpl_IDZ; omega).
rewrite <- Heq.
replace (IDR k / Z2R b * vx)%R with ((IDR k *vx )/ Z2R b)%R by field_IDR.
apply Rle_div_1; try IDR_to_Z.
replace (Z2R b) with (Z2R b*1) by ring.
replace (IDR (-Dbase)%D) with (-(Z2R b*1)) by (simpl_IDR; ring).
apply Rmult_le_compat_sym; intuition; IDR_to_Z.

assert (Heq: (IDR k/ Z2R b *vx = (IDR (k+Dbase)%D * vx +IDR 0%D)/ Z2R b - vx)%R); 
  [field_IDR | rewrite Heq].
apply Ssub_correct; (assumption|| idtac).
apply mult_div_base_correct; (assumption||simpl_IDZ; omega).
rewrite <- Heq.
replace (IDR k / Z2R b * vx)%R with ((IDR k *vx )/ Z2R b)%R by field_IDR.
apply Rle_div_1; try IDR_to_Z.
replace (Z2R b) with (Z2R b*1) by ring.
replace (IDR (-Dbase)%D) with (-(Z2R b*1)) by (simpl_IDR; ring).
apply Rmult_le_compat_sym; intuition; IDR_to_Z.

replace (IDR k / Z2R b * vx)%R with ((IDR k * vx +IDR 0%D)/Z2R b)%R by field_IDR.
apply mult_div_base_correct; (assumption||simpl_IDZ; omega).
Qed.

Theorem div_digit_aux_correct : 
  forall k r x vx,
    (2 <= IDZ k <= b)%Z ->
    (0 <= IDZ r < IDZ k)%Z -> 
    Srepresents x vx -> 
    Srepresents (div_digit_aux k r x) ((vx + IDR r)/ IDR k).  
Proof.
cofix.
intros k r x vx Hk Hr Hrep.
str_unfold (div_digit_aux k r x).
simpl; destruct x as [d x']; fold_decomp.
Ddiv_tac q r' U1 U2.
inversion Hrep; subst.
replace (((IDR d + r0) / Z2R b + IDR r) / IDR k) with 
  ((IDR q + (r0 + IDR r') / IDR k) / Z2R b).
constructor.
apply div_digit_aux_correct; assumption.
apply Rle_div_1; [IDR_to_Z|].
assert (U3 : - IDR k +1 <= IDR r' <=  IDR k - 1) by (split; IDR_to_Z).
clear div_digit_aux_correct.
dec_tac; split; fourier.
split; (apply Zmult_lt_reg_r with (IDZ k); [omega|]);
  replace (IDZ q * IDZ k)%Z with (IDZ k*IDZ q)%Z by ring; 
apply Zplus_lt_reg_r with (IDZ r'); rewrite <- U1; simpl_IDZ.
rewrite Zmult_opp_comm.
assert (G1 : (b*-IDZ k <= b*-2)%Z).
apply Zmult_le_compat_l; omega.
assert (G2 : (0 <= IDZ r * b)%Z) by ZR0compat; omega.

apply Zlt_le_trans with ((1+IDZ r)* b)%Z.
rewrite Zmult_plus_distr_l; omega.
apply Zle_trans with (IDZ k*b)%Z.
apply Zmult_le_compat_r; omega.
rewrite Zmult_comm; omega.

field_simplify_eq; [|split; IDR_to_Z].
replace (r0 + Z2R b * IDR r + IDR d) with 
  (r0 + IDR (d + r * Dbase)%D) by (simpl_IDR; ring).
unfold IDR; simpl_IDZ; rewrite U1; simpl_Z2R; ring.
Qed.

Theorem div_digit_correct :
  forall k x vx,
    (IDZ k <> 0 )%Z ->
    (-b <= IDZ k <= b)%Z ->
    Srepresents x vx -> 
    Srepresents (x / k)%S (vx/ IDR k).  
Proof.
intros k x vx Hk1 Hk2 Hrepx.
assert (Hvx : -1 <= vx <= 1).
 apply Srepresents_interval with (1:= Hrepx).
assert (Srepresents (Sopp x) (-vx)).
 apply Sopp_correct; assumption || simpl_IDZ; omega.

unfold div_digit, Dle_bool.

compare_gt_tac G1; (compare_gt_tac G2; [compare_gt_tac G3| ]).

replace (vx / IDR k) with (-vx) ; [assumption|].
replace (IDR k) with (-1) by IDR_to_Z; field.

replace (vx / IDR k) with ((-vx + IDR 0%D)/ IDR (-k)%D) by field_IDR.
apply div_digit_aux_correct; assumption || simpl_IDZ; omega.

replace (vx / IDR k) with ((IDR 0%D+-vx)/ Z2R b); [|
  replace (IDR k) with (-Z2R b) by IDR_to_Z; field_IDR].
constructor; assumption || (simpl_IDZ; omega) || dec_tac; split; fourier.

replace (vx / IDR k) with vx ; [assumption|].
replace (IDR k) with 1 by IDR_to_Z; field.

replace (vx / IDR k) with ((vx + IDR 0%D)/ IDR k) by field_IDR.
apply div_digit_aux_correct; assumption || simpl_IDZ; omega.

replace (IDR k) with (Z2R b) by IDR_to_Z.
rep_constr; [constructor; assumption ||  simpl_IDZ; omega| field_IDR].
Qed.

End MULT_DIGIT_CORRECT.