Require Export digit_big_num.
Require Export Streams.
Module EXACT_REAL_DEFINITIONS(Nums : NUMS).

Export Nums.

Infix "::" := Cons (at level 60, right associativity) :stream_scope.
Delimit Scope stream_scope with S.

Open Scope stream_scope.

Definition ER : Set :=  (nat * Stream digit)%type.

CoFixpoint repeat_digit (d : digit) : 
  Stream digit := d::repeat_digit d.

Definition Szero : Stream digit := 
  repeat_digit 0%D.

Definition Sone : Stream digit :=
  repeat_digit Dmax.

Definition Sminus_one : Stream digit := 
  repeat_digit (-Dmax)%D.

(*computes -x*)
CoFixpoint Sopp (x: Stream digit) : Stream digit :=
  match x with 
    | x1::x' => (-x1)%D::Sopp x'
  end.

(*computes |x| *)
CoFixpoint Sabs (x: Stream digit) := 
  (match x with 
    | x1::x' => 
      match 0 ?= x1 with 
        | Eq => 0::Sabs x'
        | Lt => x
        | Gt => Sopp x
      end
  end)%D.

(* if Scompare x y n = Lt then x < y *)
(* if Scompare x y n = Gt then x > y *)
(* if Scompare x y n = Eq then |x - y| <= 2 / b^n *)
Fixpoint Scompare (x y : Stream digit) (n:nat) {struct n} : comparison := 
 (match n with 
    | O => Eq
    | S p => 
      match x,y with 
	| x1::x', y1::y'  => 
	  match x1 ?= y1 with 
	    (* (x1+x')/b ?= (x1+y')/b  =  x' ?= y'*)
	    | Eq => Scompare x' y' p
	    | Lt => 
	      match x1 + 1 ?= y1 - 1 with 
		| Lt => Lt (* x1 +2 < y1 => x < y*)
 		| Eq =>    (* (x1+x')/b ?= (x1+2+y')/b  =  x' ?= y' + 2*)
		  match x' with
		    | x2::x'' => 
		      match x2 ?= Dmax with 
			| Eq => 
			  match y' with 
			    | y2::y'' => 
			      match y2 ?= - Dmax with 
				| Eq => Scompare (-(1)::x'') (1::y'') p
				| _ => Lt
			      end
			  end
			| _ => Lt
		      end
		  end
		| Gt => (*y1 = x1+1*)
		  match x' with
		    | x2::x'' => 
		      match x2 ?= - (1) with 
			| Lt => Lt
			| _ => 
			  match y' with 
			    | y2::y'' => 
			      match y2 ?= 1 with
				| Gt => Lt
				| _ => Scompare (x2-2::x'') (y2+Dmax-1::y'') p
			      end
			  end
		      end
		  end
	      end
	    | Gt => 
	      match x1 - 1 ?= y1 + 1 with 
		| Gt => Gt (* x1 < y1 +2 => x < y*)
 		| Eq =>    (* (x1+x')/b ?= (x1+2+y')/b  =  x' ?= y' + 2*)
		  match x' with
		    | x2::x'' => 
		      match x2 ?= - Dmax with 
			| Eq => 
			  match y' with 
			    | y2::y'' => 
			      match y2 ?= Dmax with 
				| Eq => Scompare (1::x'') (-(1)::y'') p
				| _ => Gt
			      end
			  end
			| _ => Gt
		      end
		  end
		| Lt => (*y1 = x1+1*)
		  match x' with
		    | x2::x'' => 
		      match x2 ?= 1 with 
			| Gt => Gt
			| _ => 
			  match y' with 
			    | y2::y'' => 
			      match y2 ?= - (1) with
				| Lt => Gt
				| _ => Scompare (x2+Dmax-1::x'') (y2-2::y'') p
			      end
			  end
		      end
		  end
	      end
	  end
      end
  end)%D.

(* if Scompare2 x y n = Lt then x <= y *)
(* if Scompare2 x y n = Gt then x >= y *)
(* if Scompare2 x y n = Eq then |x - y| < 2 / b^n *)
Fixpoint Scompare2 (x y : Stream digit) (n:nat) {struct n} : comparison := 
 (match n with 
    | O => Eq
    | S p => 
      match x,y with 
	| x1::x', y1::y'  => 
	  match x1 ?= y1 with 
	    (* (x1+x')/b ?= (x1+y')/b  =  x' ?= y'*)
	    | Eq => Scompare2 x' y' p
	    | Lt => 
	      match x1 + 1 ?= y1  with 
		| Eq =>    (* (x1+x')/b ?= (x1+2+y')/b  =  x' ?= y' + 2*)
		  match x' with
		    | x2::x'' => 
		      match x2 ?= 0 with 
			| Lt => Lt
			| _ => 
			  match y' with 
			    | y2::y'' => 
			      match y2 ?= 0 with
				| Gt => Lt
				| _ => Scompare2 (x2-1::x'') (y2+Dmax::y'') p
			      end
			  end
		      end
		  end
		| _ => Lt
	      end
	    | Gt => 
	      match x1 ?= y1 + 1 with 
		| Eq => (*y1 = x1+1*)
		  match x' with
		    | x2::x'' => 
		      match x2 ?= 0 with 
			| Gt => Gt
			| _ => 
			  match y' with 
			    | y2::y'' => 
			      match y2 ?= 0 with
				| Lt => Gt
				| _ => Scompare2 (x2+Dmax::x'') (y2-1::y'') p
			      end
			  end
		      end
		  end
		| _ => Gt
	      end
	  end
      end
  end)%D.

(* Computes min the minimum of s1 s2*)
CoFixpoint Smin (x y : Stream digit) : Stream digit := 
  (match x,y with 
     | x1::x', y1::y'  => 
       match x1 ?= y1 with 
	 | Eq => x1::Smin x' y'
	 | Lt => 
	   match x1 + 1 ?= y1  with 
	     | Eq => 
	       match x' with
		 | x2::x'' => 
		   match x2 ?= 0 with 
		     | Lt => x
		     | Eq => 
		       match y' with 
			 | y2::y'' => 
			   match y2 ?= -Dmax with
			     | Eq => x1::Smin x' (1::y'')
			     | _ => x
			   end
		       end
		     | Gt =>
		       match y' with 
			 | y2::y'' => 
			   match y2 ?= 0 with
			     | Gt => x
			     | _ => y1::Smin (x2-Dbase::x'') y'
			   end
		       end
		   end
	       end
	     | _ => x
	   end
	 | Gt => 
	   match x1 ?= y1 + 1 with 
	     | Eq => (*y1 = x1+1*)
	       match x' with
		 | x2::x'' => 
		   match x2 ?= 0 with 
		     | Gt => y
		     | Eq => 
		       match y' with 
			 | y2::y'' => 
			   match y2 ?= Dmax with
			     | Eq => x1::Smin x' (-(1)::y'')
			     | _ => y
			   end
		       end
		     | Lt => 
		       match y' with 
			 | y2::y'' => 
			   match y2 ?= 0 with
			     | Lt => y
			     | _ => y1::Smin (x2 + Dbase::x'') y'
			   end
		       end
		   end
	       end
	     | _ => y
	   end
       end
   end)%D.

CoFixpoint Smax (x y : Stream digit) : Stream digit := 
  (match x,y with 
     | x1::x', y1::y'  => 
       match x1 ?= y1 with 
	 | Eq => x1::Smax x' y'
	 | Lt => 
	   match x1 + 1 ?= y1  with 
	     | Eq => 
	       match y' with
		 | y2::y'' => 
		   match y2 ?= 0 with 
		     | Gt => y
		     | Eq => 
		       match x' with 
			 | x2::x'' => 
			   match x2 ?= Dmax with
			     | Eq => y1::Smax (-(1)::x'') y'
			     | _ => y
			   end
		       end
		     | Lt =>
		       match x' with 
			 | x2::x'' => 
			   match x2 ?= 0 with
			     | Lt => y
			     | _ => x1::Smax x' (y2+Dbase::y'')
			   end
		       end
		   end
	       end
	     | _ => y
	   end
	 | Gt => 
	   match x1 ?= y1 + 1 with 
	     | Eq => 
	       match x' with
		 | x2::x'' => 
		   match x2 ?= 0 with 
		     | Gt => x
		     | Eq => 
		       match y' with 
			 | y2::y'' => 
			   match y2 ?= Dmax with
			     | Eq => x1::Smax x' (-(1)::y'')
			     | _ => x
			   end
		       end
		     | Lt => 
		       match y' with 
			 | y2::y'' => 
			   match y2 ?= 0 with
			     | Lt => x
			     | _ =>  y1::Smax (x2+Dbase::x'') y'
			   end
		       end
		   end
	       end
	     | _ => x
	   end
       end
   end)%D.

Definition Dle_bool x y := 
  (match x ?= y with 
     | Gt => false
     | _ => true
   end)%D.

Infix "<=" := Dle_bool : digit_scope.

(*multiplication of a Stream with the base*)
CoFixpoint multb (x : Stream digit) : Stream digit :=
  (match x with 
    | x1::x' => 
      match x1 ?= 1 with 
        | Eq => 
          match x' with 
	    | x2::x'' =>
              match x2 ?= 0 with 
                | Eq => Dmax::multb (1::x'')
                | Gt => Sone 
                | Lt => x2+Dmax+1::x''
              end
          end
        | Gt => Sone 
        | Lt => 
          match x1 ?= -(1) with 
            | Eq =>  
              match x' with 
	        | x2::x'' =>
                  match x2 ?= 0 with 
                    | Eq => -Dmax::multb (-(1)::x'')
                    | Gt => x2-Dmax-1::x''
                    | Lt => Sminus_one 
                  end
              end
            | Gt => x'
            | Lt => Sminus_one 
          end
      end
   end)%D.

Definition find_sum_remainder (x1 y1 : digit) : (digit * digit) :=
  (let s:= x1 + y1 in 
    match s ?= Dmax with 
      | Lt => 
        match s ?= -Dmax with 
          | Gt => (0, s)
          | _ => (-(1), s + Dbase)
        end
      | _ => (1, s - Dbase)
    end)%D.

(*computes (x+y+r)/b*)
CoFixpoint sum_div_base (x y: Stream digit) (r : digit) :
  Stream digit := 
  match (x, y) with 
    | (x1::x', y1::y') =>
      let (q, r') := find_sum_remainder x1 y1 in 
        (q+r)%D::sum_div_base x' y' r'
  end.

(*computes x+y when -1 <= x+y <= 1*)
Definition Sadd (x y: Stream digit) : Stream digit:=
  multb (sum_div_base x y 0)%D.

Infix "+" := Sadd :stream_scope.

(*computes (x-y+r)/b*)
CoFixpoint sub_div_base (x y: Stream digit) (r : digit) : Stream digit := 
  match (x, y) with 
    | (x1::x', y1::y') => 
     let (q, r') := find_sum_remainder x1 (-y1)%D in 
        (q+r)%D::sub_div_base x' y' r'
  end.

(*computes x-y when -1 <= x-y <= 1*)
Definition Ssub (x y: Stream digit) : Stream digit:=
  multb (sub_div_base x y 0%D).

Infix "-" := Ssub :stream_scope.

Definition find_mult_remainder (k x1 : digit) : (digit * digit) :=
  (let (q,r') := Ddiv_eucl (k*x1) Dbase in 
    match  r' +  r' ?= Dbase with 
      | Gt => (q+1, r'-Dbase)
      | _ =>
        match q+q ?= -Dbase with 
          | Gt => (q, r')
          | _ => (q+1, r'-Dbase)
        end
    end)%D.

(*computes (k*x+r)/b with (-(b+1)<=2*k<=b+1 *)
CoFixpoint mult_div_base (k r: digit) (x : Stream digit) : Stream digit := 
  match x with 
    | x1::x' => let (q,r') := find_mult_remainder k x1 in 
      (q+r)%D::mult_div_base k r' x'
  end.

(* (*computes (k*x+r)/b with (-(b+1)<=2*k<=b+1 *) *)
(* CoFixpoint mult_div_base (k r: digit) (x : Stream digit) : Stream digit :=  *)
(*   match x with  *)
(*     | x1::x' => let (q,r') := Ddiv_eucl (k*x1)%D Dbase in  *)
(*       if (r'+r' <= Dbase)%D *)
(* 	then if (q+q<=-Dbase)%D *)
(* 	  then (q+1+r)%D::mult_div_base k (r'-Dbase)%D x' *)
(* 	  else (q+r)%D::mult_div_base k r' x' *)
(* 	else (q+1+r)%D::mult_div_base k (r'-Dbase)%D x' *)
(*   end. *)

(*computes (k*x)/Dbase with k a digit of b*)
Definition mult_digit (k: digit) (x : Stream digit) : Stream digit :=  
  if (k+k <= Dbase+1)%D
    then if (-Dbase-1<= k+k)%D
      then mult_div_base k 0%D x
      else (mult_div_base (k+Dbase)%D 0%D x) - x
    else (mult_div_base (k-Dbase)%D 0%D x) +  x.

(*Infix "*D" := mult_digit :stream_scope.*)

(*computes (x+r)/k with 2 <=k<=b and 0 <= r < k *)
CoFixpoint div_digit_aux (k r: digit) (x : Stream digit) : Stream digit := 
  match x with 
    | x1::x' => let (q,r') := Ddiv_eucl (x1+r*Dbase)%D k in
      q::div_digit_aux k r' x'
  end.

(*computes x/k with k a digit of b*)
Definition div_digit (x : Stream digit) (k: digit) : Stream digit :=  
  (if 0 <= k
    then if Dbase <= k
      then 0::x
      else if 2<= k
	then div_digit_aux k 0 x
	else x
    else if k <= -Dbase 
      then 0::Sopp x
      else if k <= -(2)
	then div_digit_aux (-k) 0 (Sopp x)
	else Sopp x)%D.

Infix "/" := div_digit :stream_scope.

(* computes a/c with 0 <= a < c*)
CoFixpoint rat_to_stream_pos (a c: big) : Stream digit:= 
  let (k,r) := Bdiv_eucl (a*Bbase)%B c in
    (digit_of_big k)::rat_to_stream_pos r c.

(* computes -a/c with a >0 and c >0 *)
CoFixpoint rat_to_stream_neg (a c: big) : Stream digit:= 
  let (k,r) := Bdiv_eucl (a*Bbase)%B c in
    (digit_of_big (-k)%B)::rat_to_stream_neg r c.

(* constructs the infinite Stream representing a rational number.
   This function works even if the rational number is not between
   0 and 1, but the value is meaningless. *)
Definition Srat_to_stream (a c : big) : Stream digit:= 
  if (0 <= c)%B
    then if (0 <= a)%B 
      then if (c <= a)%B
	then Sone 
	else rat_to_stream_pos a c
      else if (c <= -a)%B
	then Sminus_one 
	else rat_to_stream_neg (-a)%B c
    else if (a <= 0)%B
      then if (a <= c)%B
	then Sone 
	else rat_to_stream_pos (-a)%B (-c)%B
      else if (-a <= c)%B
	then Sminus_one 
	else rat_to_stream_neg a (-c)%B.

Fixpoint bounds_aux (n: nat) (acc: big) (x: Stream digit) {struct n} : big :=
  match n with 
    O => acc
    | S p => 
      match x with 
	k::x'=> bounds_aux p (acc*Bbase+big_of_digit k)%B x'
      end
  end.

(* returns an approximation a such that (a-1)/b^n < x < (a+1)/b^n *)
Definition bounds (n : nat) (x: Stream digit) : big := 
  bounds_aux n 0%B x.

(*************************)
(*******  SERIES  ********)
(*************************)

Definition make_digit (r : Stream digit) : Stream digit :=
  match r with
    | k1::k2::r' =>
      if (-Dbase <= k2+k2)%D
	then if (k2+k2 <= Dbase)%D
	  then k1::k2::r'
	  else if (Dmax <= k1)%D
	    then  k1::k2::r'
	    else (k1+1)%D::(k2-Dbase)%D::r'
	else if (k1 <= -Dmax)%D
	  then k1::k2::r'
	  else (k1-1)%D::(k2+Dbase)%D::r'
  end.

Definition series_body (f: Stream digit -> Stream digit) 
  (r : Stream digit) : Stream digit := 
  match (make_digit r) with 
    k::r'=> k::f r'
  end. 

(* computes x + u*v*) 
CoFixpoint mult_series (u v r : Stream digit) : Stream digit :=
  match u with
    k1::u' => 
    series_body (mult_series u' v) (r + (mult_digit k1 v))
  end.

(*Computes x*y *)
Definition Smult (x y: Stream digit): Stream digit :=
  multb (multb (mult_series x (0::0::y) Szero))%D.

Infix "*" := Smult :stream_scope.

(*Inverse*)


Fixpoint inv_step (y z k : big) (x xn sx : Stream digit) (n: nat) :=
  match n with 
    | O => (z, xn, sx)
    | S p => 
      if (y <= z)%B 
        then (z, xn, sx)
        else inv_step y (z*k)%B k x (xn * x)%S (xn + sx)%S p 
 end.

CoFixpoint inv_series (y z k :big)  (n: nat) (x xn r: Stream digit) :=
  match inv_step y z k x xn Szero n with 
    | (z', xn', sx) => 
      series_body (inv_series (y*Bbase)%B z' k n x (multb xn')) (r +sx)%S
  end.

(* computes l such that  k^(n+l) < b <=  k^(n+l+1) *)
Fixpoint log_aux (n : nat) (k kn: digit) : nat :=
  (match n with 
    O => O
    | S p => 
      let kn':= k*kn in 
	if  Dbase <= kn' 
	  then O
	  else S (log_aux p k kn') 
  end)%D.

(*computes l such that k^l < b < k^(l+1)*)
Definition log (k : digit) :=
  log_aux (IDN (Dlog2 Dbase))%nat k 1%D.

Definition inv_init x k := 
  let k_big:= (big_of_digit k) in
    if (2 * Bbase <= (Bbase-4) * (k_big-1))%B
      then inv_series (2*Bbase*Bbase)%B (Bbase-4)%B 
        k_big (log k + 1)%nat x (Sone / k)%S Szero
      else if (2 * Bbase <= (Bbase-4) * (k_big-1)* k_big)%B
        then (Sone /k + inv_series (2*Bbase*Bbase)%B (k_big *(Bbase-4))%B 
          k_big (log k + 1)%nat x (x / k)%S Szero)%S
        else if (2 * Bbase <= (Bbase-4) * (k_big-1)* k_big* k_big)%B
          then ((Sone /k + x /k)+ inv_series (2*Bbase*Bbase)%B
            (k_big *k_big *(Bbase-4))%B 
            k_big (log k + 1)%nat x ((x*x) / k)%S Szero)%S
          else if (2 * Bbase <= (Bbase-4) * (k_big-1)* k_big* k_big* k_big)%B
          then ((Sone /k + x /k + (x*x)/k)+ inv_series (2*Bbase*Bbase)%B
            (k_big *k_big *k_big *(Bbase-4))%B 
            k_big (log k + 1)%nat x ((x*x*x) / k)%S Szero)%S
          else ((Sone /k + x /k + (x*x)/k+(x*x*x)/k)+ inv_series (2*Bbase*Bbase)%B
            (k_big *k_big *k_big *k_big *(Bbase-4))%B 
            k_big (log k + 1)%nat x ((x*x*x*x) / k)%S Szero)%S.

(*Adds n zeros in head of the stream*)
Fixpoint str_add_zero (n : nat) (s : Stream digit) {struct n}: Stream digit :=
  match n with 
    | O => s
    | S p => str_add_zero p (0::s)%D
  end.


(*computes b^e2 / x if |x| >= 1/b^e1 *)
Fixpoint Sinverse (n : nat) (s : Stream digit) {struct n} : Stream digit :=
  (match n with 
    | O => s
    | S n => 
      match s with 
	| k::s' => 
	  match k ?= 1 with 
            | Gt => str_add_zero n (inv_init (s' / (-k))%S k)
	    | Eq => 
	      match s' with 
		k'::s'' =>
 	        match k' ?= 0 with 
                  | Gt => str_add_zero n (inv_init ((k'-Dbase::s'')%D/ (-(2)))%S 2)
                  | Eq => multb (str_add_zero n (inv_init ( s''/ (-Dbase))%S Dbase))
                  | Lt => Sinverse n (k'+ Dbase::s'')
                end
              end
            | Lt =>
              match k ?= -(1) with 
                | Gt => Sinverse n s'
                | Lt => str_add_zero n (Sopp (inv_init (s' / (-k))%S (-k)))
                | Eq => 
                  match s' with 
		    k'::s'' =>
                    match  k' ?= 0 with 
                      | Lt => str_add_zero n (Sopp (inv_init ((k'+Dbase::s'')%D/ 2)%S 2))
                      | Eq => multb (str_add_zero n (Sopp (inv_init (s''/ Dbase)%S Dbase)))
		      | Gt => Sinverse n (k'-Dbase::s'')
		    end
                  end
              end
          end
      end
   end)%D.

(* PI *)

Definition numB (k1 k2 k3 : big) : big :=
  (k2*(8*k1+18)+k3*(3*k1+8))%B.

Definition denomB (k1 k2 k3 : big) : big :=
  (k1*(k1+2)*k2*k3)%B.

Fixpoint finite_sum_pi_aux (k1 k2 k3 p q z : big) (n: nat) {struct n} :=
  (let k1' := k1+4 in 
    let k2' := k2*16 in 
      let k3':= k3*81 in 
	match n with
	  | O => (k1', k2', k3', p, q)
	  | S m => 
	    let w := (k1'*(k1'+2))%B in
	      finite_sum_pi_aux k1' k2' k3'
	      (p*1296*w+z*numB k1' k2' k3') (q*1296*w) (z*w) m
  end)%B.

Definition finite_sum_pi (k1 k2 k3 : big) (step : nat):= 
  (finite_sum_pi_aux k1 k2 k3 (numB k1 k2 k3)
    (denomB k1 k2 k3) (k1*(k1+2)))%B (step-1)%nat.

CoFixpoint pi_series (k1 k2 k3 y : big) (step : nat) (r : Stream digit) : Stream digit :=
  match finite_sum_pi k1 k2 k3 step with
    (k1', k2', k3', p, q) =>
    series_body (pi_series k1' k2' k3'
      (Bbase*y)%B step) ( r + (Srat_to_stream (y*p)%B q))
  end.

Definition find_step :=
  let (q,_) := (Ddiv_eucl (Dlog2 (Dmax)) 4)%D in 
    S (IDN q).

Definition pi_div4 :=
  let step := find_step in
    let (aux, y) :=  finite_sum_pi 1%B 8%B 27%B (S step) in
      let (aux1, x) := aux in let (aux2, k3) := aux1 in
        let (k1, k2) := aux2 in
          (pi_series k1 k2 k3 1%B step (Srat_to_stream x y)).

(*****************
 ******  R  ******
 *****************)

(* Try to remove e zeros from the head of s. 
   Returns also the number of zeros not removed *)
Fixpoint str_del_zero (e : nat) (s : Stream digit) {struct e} : ER :=
  match e with 
    | O => (e,s)
    | (S p) => 
      match s with 
	| k::s' => 
	  match (k ?= 0)%D with 
            | Eq => str_del_zero p s'
            | _ => (e,s)
	  end
      end
  end.


(*adds zero in the begining of on Stream such that the 2 streams have the same exponant*)
Definition exp_norm (x y : ER) : ER * Stream digit :=
  let (e1, x'):= x in 
    let (e2, y'):= y in
      if (le_lt_dec e1 e2) 
	then (e2, (str_add_zero (e2-e1) x'), y')
	else (e1, x', (str_add_zero (e1-e2) y')).

(*computes x + y*)
Definition add (x y : ER) : ER :=
  match exp_norm x y with 
    | (e, x', y') => 
      (e+1, sum_div_base x' y' 0%D)%nat
  end.

(*computes x - y*)
Definition sub (x y : ER) : ER :=
  match exp_norm x y with 
    | (e, x', y') => 
      (e+1, sub_div_base x' y' 0%D)%nat
  end.

(*Computes x*y *)
Definition mult (x y : ER) : ER :=
  let (e1, x') := x in 
    let (e2, y'):= y in 
      ((e1+e2+2)%nat, mult_series x' (0::0::y')%D  Szero).

(*Computes - x*)
Definition opp (x : ER) : ER :=
  let (e, x') := x in
    (e, Sopp x').

Definition zero :=
  (0%nat, Szero).

Definition one :=
  (0%nat, Sone).

Definition pi := 
  (1%nat, mult_digit 4%D (pi_div4)).

Infix "+" := add :exact_real_scope.
Infix "-" := sub :exact_real_scope.
Infix "*" := mult :exact_real_scope.

Delimit Scope exact_real_scope with ER.

(* if compare x y n = Lt then x < y *)
(* if compare x y n = Gt then x > y *)
(* if compare x y n = Eq then |x - y| <= 2 / b^n *)
Definition compare (x y : ER) (n:nat) := 
  match exp_norm x y with 
    | (e , x', y') => 
      Scompare x' y' (n+e)
  end.
  
(* if compare2 x y n = Lt then x <= y *)
(* if compare2 x y n = Gt then x >= y *)
(* if compare2 x y n = Eq then |x - y| <= 2 / b^n *)
Definition compare2 (x y : ER) (n:nat) := 
  match exp_norm x y with 
    | (e , x', y') => 
      Scompare2 x' y' (n+e)
  end.

(*Computes min (x, y) *)
Definition min (x y : ER) := 
  match exp_norm x y with 
    | (e , x', y') =>  (e, Smin x' y')
  end.

(*Computes max (x, y) *)
Definition max (x y : ER) := 
  match exp_norm x y with 
    | (e , x', y') =>  (e, Smax x' y')
  end.

(*Computes the absolute value*)
Definition abs (x: ER) := 
  match x with 
    (e, x) => (e, Sabs x)
  end.

(*Computes (y', l) st y = y' * b^l and x <= y' *)
Fixpoint Blog_aux (x y: big) (l n: nat) {struct n} := 
  match n with 
    | O => (y, l)
    | S p =>
      if (x <= y)%B 
	then (y, l)
	else Blog_aux x (Bbase*y)%B (S l) p
  end.

Definition rat_to_stream (x y : big) : (ER) :=
  (if 1 <= x 
    then if 0 <= y
      then let (q, r) := Bdiv_eucl x y in
	let n := (Zabs_nat (IBZ q)) in 
	  let (y', l) := Blog_aux (x+1) y O n in
	    (l, Srat_to_stream x y')
      else let (q, r) := Bdiv_eucl x (-y) in
	let n := (Zabs_nat (IBZ q)) in 
	  let (y', l) := Blog_aux (x+1) (-y)%B O n in
	    (l, Srat_to_stream x (- y'))
    else if 0 <= x 
      then zero
      else if 0 <= y
	then let (q, r) := Bdiv_eucl (-x) y in
	  let n := (Zabs_nat (IBZ q)) in 
	    let (y', l) := Blog_aux (- x +1) y O n in
	      (l, Srat_to_stream x y')
	else let (q, r) := Bdiv_eucl (-x) (-y) in
	  let n := (Zabs_nat (IBZ q)) in 
	    let (y', l) := Blog_aux (-x+1) (-y)%B O n in
	      (l, Srat_to_stream x (- y')))%B.


(*Computes the inverse of x if its |x|>= b^n something else otherwise*)
Definition inverse (x : ER) (n : nat) :=
  match x with 
    (e, x') => (n, (Sinverse (n+e) x'))
  end.

End EXACT_REAL_DEFINITIONS.
