Require Import proof_utils.

Theorem sum_f_single : 
  forall An k, 
    sum_f k k An = An k.
Proof.
intros An k.
unfold sum_f.
rewrite <- minus_n_n.
simpl; ring.
Qed.

Theorem sum_f_step : 
  forall (An : nat -> R) (k1 k2 : nat),
    (k1 < k2)%nat ->
    sum_f k1 k2 An = An k1 + sum_f (k1+1) k2 An.
Proof.
intros An k1 k2 H1.
unfold sum_f.
rewrite decomp_sum; try omega.
apply Rplus_eq_compat_l.
replace (k2 - (k1 + 1))%nat with (pred (k2 - k1)). 
apply sum_eq.
intros i H2.
replace (S i +k1)%nat with (i + (k1 +1))%nat by ring; reflexivity. 
omega.
Qed.

Theorem sum_f_step_r : 
  forall (An : nat -> R) (k1 k2 : nat),
    (k1 <= k2)%nat ->
    sum_f k1 (k2+1) An = sum_f k1 k2 An + An (k2+1)%nat.
Proof.
intros An k1 k2 H1.
assert (R1: (k2=k1 + (k2-k1))%nat) by omega.
generalize H1; clear H1.
rewrite R1; clear R1.
generalize (k2-k1)%nat.
intros n.
generalize An k1.
clear An k1 k2.
induction n.
intros An k1 H1.
rewrite <- plus_n_O.
rewrite sum_f_step; try omega.
repeat rewrite sum_f_single; reflexivity.

intros An k1 H1.
repeat rewrite (sum_f_step An k1); try omega.
replace (k1 + S n + 1)%nat with ((k1+1) + n + 1)%nat by ring.
rewrite IHn; try omega.
replace (k1 + S n)%nat with (k1 + 1 + n)%nat; ring.
Qed.

Theorem sum_f_decomp : 
  forall (An : nat -> R) (k1 k2 k3: nat),
    (k1 <= k2 < k3)%nat ->
    sum_f k1 k3 An = sum_f k1 k2 An + sum_f (k2+1)%nat k3 An.
Proof.
intros An k1 k2 k3 H1.
assert (R1: (k2=k1 + (k2-k1))%nat) by omega.
generalize H1; clear H1.
rewrite R1; clear R1.
generalize (k2-k1)%nat.
intros n.
generalize An k1 k3.
clear An k1 k2 k3.
induction n.
intros An k1 k3 H1.
rewrite <- plus_n_O.
rewrite sum_f_single.
apply sum_f_step; omega.

intros An k1 k3 H1.
repeat rewrite (sum_f_step An k1); try omega.
rewrite IHn; try omega.
replace (k1 + S n + 1)%nat with (k1+1 + n + 1)%nat by ring.
replace (k1 + S n)%nat with (k1 + 1 + n)%nat; ring.
Qed.

Theorem decomp_sum_S:
  forall (An : nat -> R) (N : nat),
    sum_f_R0 An (S N) = An 0%nat + sum_f_R0 (fun i : nat => An (S i)) N.
Proof.
intros An N.
induction N.
reflexivity.
pattern (S N) at 1.
simpl sum_f_R0.
lazy beta.
rewrite IHN; ring.
Qed.

Theorem infinit_sum_cv :
  forall An l, infinit_sum An l -> Un_cv (fun n => sum_f_R0 An n) l.
unfold infinit_sum, Un_cv; auto.
Qed.

Theorem sum_cv_infinit_sum :
  forall An l, Un_cv (fun n => sum_f_R0 An n) l -> infinit_sum An l.
unfold infinit_sum, Un_cv; auto.
Qed.

Theorem infinit_sum_ext :
  forall (An Bn : nat -> Rdefinitions.R) (l:Rdefinitions.R),
  (forall n, An n = Bn n) ->
  infinit_sum An l -> infinit_sum Bn l.
Proof.
unfold infinit_sum.
intros An Bn l Hext  Hinf.

intros eps Hepspos; elim (Hinf eps Hepspos); intros n Hdist.
exists n; intros p Hpgen; apply Rle_lt_trans with (2:=Hdist p Hpgen).
rewrite (sum_eq An Bn p);auto with real.
Qed.

Theorem infinit_sum_step :
  forall (An : nat -> Rdefinitions.R) (l:Rdefinitions.R),
   infinit_sum An l -> infinit_sum (fun i => An (i+1)%nat) (l-An 0%nat)%R.
Proof.
unfold infinit_sum.
intros An l H.
intros eps Hepspos; elim (H eps Hepspos); intros n Hdist.
exists (n+1)%nat; intros p Hpgen; generalize (Hdist (p+1)%nat).
rewrite (tech2 An 0 (p+1)%nat).
simpl sum_f_R0.
unfold R_dist.
replace (sum_f_R0 (fun i : nat => An (S i)) (p + 1 - 1)%nat) with
  (sum_f_R0 (fun i => An (i+1)%nat) p).
replace (p + 1 - 1)%nat with p.
match goal with |- (_ -> (Rabs ?x < eps)%R) -> (Rabs ?y < eps)%R =>
 replace x with y;[intros H'; apply H'; clear H'| ring]
end.
omega.
rewrite plus_comm; rewrite minus_plus;auto.
rewrite plus_comm; rewrite minus_plus.
apply (sum_eq (fun i => An (i+1)%nat) (fun i => An (S i)) p).
intros; replace (S i) with (i+1)%nat; auto; ring.
omega.
Qed.

Theorem infinit_sum_bound :
  forall An Bn l1 l2, 
   (forall n, (An n <= Bn n)%R) ->
   infinit_sum An l1 -> infinit_sum Bn l2 -> (l1 <= l2)%R.
Proof.
intros An Bn l1 l2 Hextle Hinf1 Hinf2.
apply Rle_cv_lim with (fun i => sum_f_R0 An i)(fun i => sum_f_R0 Bn i).
intros n; apply sum_Rle.
intros; apply Hextle.
apply infinit_sum_cv;assumption.
apply infinit_sum_cv;assumption.
Qed.

Theorem bound_infinit_sum :
  forall An Bn l,
    (forall n, 0 <= An n <= Bn n)%R ->
    infinit_sum Bn l ->
    (exists l2, infinit_sum An l2 /\ (0 <= l2 <= l)%R).
intros An Bn l H Hinf.
assert (Hcv1 := infinit_sum_cv Bn l Hinf).

assert (Hcv := Rseries_CV_comp An Bn
               H (existT (fun l => Un_cv (fun i => sum_f_R0 Bn i) l) l Hcv1)).
destruct Hcv as [l2 Hcv].
exists l2; split.
apply sum_cv_infinit_sum; assumption.
split.
apply infinit_sum_bound with (fun i:nat => 0%R) (fun i => An i).
intros n; elim (H n);intros; assumption.
intros eps; exists 0%nat.
intros.
rewrite sum_eq_R0.
rewrite R_dist_eq.
fourier.
auto.
apply infinit_sum_ext with An.
auto.
apply sum_cv_infinit_sum; auto.
apply infinit_sum_bound with An Bn.
intros n; elim (H n); auto.
apply sum_cv_infinit_sum;auto.
apply sum_cv_infinit_sum;auto.
Qed.

Theorem scal_infinit_sum : 
  forall (An : nat -> Rdefinitions.R) (x r: Rdefinitions.R),
    infinit_sum An r ->
    (x <> 0)%R ->
    infinit_sum  (fun i : nat => (An i * x)%R) (x*r).
Proof.
intros An x r.
unfold infinit_sum.
intros Hinf.
intros HRabs eps Heps_pos.
assert (Heps'_pos: (eps / Rabs x>0)%R).
unfold Rdiv.
assert (0<eps* / Rabs x)%R.
apply Rlt_mult_inv_pos.
auto with real.
apply Rabs_pos_lt.
assumption.
auto with real.
generalize (Hinf (eps/ (Rabs x))%R Heps'_pos).
clear Hinf.
intros [N0 Hinf].
exists N0.
intros n HngeN0.
generalize (Hinf n HngeN0).
clear Hinf.
unfold R_dist.
unfold Rdiv.
intros Hinf.
rewrite <- scal_sum.
rewrite <- Rmult_minus_distr_l.
rewrite Rabs_mult.
rewrite Rmult_comm.
apply Rmult_lt_reg_l with (/Rabs x).
rewrite <- Rabs_Rinv.
apply Rabs_pos_lt.
auto with real.
assumption.
rewrite Rmult_comm.
rewrite Rinv_r_simpl_l.
rewrite Rmult_comm.
assumption.
apply Rabs_no_R0.
assumption.
Qed.

Theorem infinit_sum_step_k:
  forall (An : nat -> R) (l : R) n,
    infinit_sum An l ->
    infinit_sum (fun i : nat => An (i + S n)%nat) 
    (l - sum_f_R0 An n).
Proof.
induction n.
intros Hinf.
unfold sum_f_R0.
apply infinit_sum_step.
exact Hinf.
intros Hinf.
simpl.
generalize (IHn Hinf).
intros U.
generalize (infinit_sum_step _ _ U).
simpl.
replace (l - (sum_f_R0 An n + An (S n)))%R with 
  ((l-sum_f_R0 An n) - An (S n))%R; [idtac|ring].
apply infinit_sum_ext.
intros n0.
replace (n0 +1 + S n)%nat with (n0 +S (S n))%nat; [reflexivity| ring].
Qed.

Theorem infinit_sum_0 : 
  infinit_sum (fun _ : nat => 0) 0.
Proof.
unfold infinit_sum.
intros eps Heps_pos. 
exists O.
intros n Hnpos.
rewrite sum_cte.
rewrite Rmult_0_l.
rewrite R_dist_eq; assumption.
Qed.

Theorem finite_GP :
  forall x n, 
    x <> 1%R -> 
    sum_f_R0 (fun i => x^i)%R n = ((1-x^(n+1)%nat) / (1-x))%R.
Proof.
intros x n H.
apply Rmult_eq_reg_l with (x-1)%R.
rewrite Rmult_comm.
rewrite GP_finite.
field.
auto with real.
auto with real.
Qed.

Theorem pow_lt_1_decr : 
  forall x n m, (0 < x <= 1 -> (n<=m)%nat -> x ^m <= x^n)%R.
Proof.
intros x n m [H1 H2] H3.
replace m with ((m-n)+n)%nat.
replace (x  ^ n)%R with (1* x ^ n)%R.
rewrite pow_add. 
apply Rmult_le_compat_r.
apply pow_le.
auto with real.
rewrite <- (pow1 (m-n)).
apply pow_incr.
auto with real.
auto with real.
omega.
Qed.

Theorem Rabs_div:
  forall x y : R, 
      y <> 0 ->
      Rabs (x / y) = Rabs x / Rabs y.
Proof.
intros x y Hy.
unfold Rdiv.
rewrite <- Rabs_Rinv; try assumption.
apply Rabs_mult.
Qed.

Theorem INR_Zabs_nat_eq :
  forall z, (0 <= z)%Z -> (IZR z = INR(Zabs_nat z))%R.
unfold Zabs_nat.
intros z; case z.
simpl.
intros; auto.
unfold IZR.
intros; auto.
compute.
intros x H'; elim H'; auto.
Qed.

Theorem pow_exp : forall x, 0 < x -> forall n, x^n=exp (ln x * (INR n)).
intros x Hpos.
induction n.
rewrite Rmult_0_r.
rewrite exp_0; trivial.
rewrite S_INR; simpl.
rewrite Rmult_plus_distr_l.
rewrite Rmult_1_r.
rewrite exp_plus.
rewrite exp_ln.
rewrite IHn; ring.
assumption.
Qed.

Theorem pow_lt_1 :
  forall eps, 0 < eps -> 
    forall x, 0 <= x < 1 -> 
      exists n:nat, x^n < eps.
intros eps Hpos x Hint.
elim (Rlt_dec 0 x).
intros H0ltx.
assert (Hlnxneg:ln x < 0).
rewrite <- ln_1.
apply ln_increasing;auto.
intuition.
elim (Rlt_dec 1 eps).
intros Hepslt1.
exists 1%nat; simpl; rewrite Rmult_1_r.
apply Rlt_trans with 1; intuition auto with real.
intros Hn1lteps.
elim (Rlt_dec eps 1).
intros Hepslt1.
elim (archimed (ln eps / ln x)).
exists (Zabs_nat (up (ln eps / ln x))).
rewrite pow_exp.
rewrite <- INR_Zabs_nat_eq.
apply Rlt_le_trans with (exp (ln x * (ln eps / ln x))).
apply exp_increasing.
apply Ropp_lt_cancel.
repeat rewrite <- Ropp_mult_distr_l_reverse.
apply Rmult_lt_compat_l.
replace 0 with (-0).
apply Ropp_lt_contravar.
rewrite <- ln_1.
apply ln_increasing.
assumption.
intuition.
ring.
auto with real.
unfold Rdiv; rewrite (Rmult_comm (ln eps));rewrite <- Rmult_assoc.
rewrite Rinv_r; auto with real.
rewrite Rmult_1_l; rewrite exp_ln; auto with real.
auto with real.
apply le_IZR.
apply Rlt_le.
apply Rle_lt_trans with (ln eps / ln x); auto with real.
simpl (IZR 0).
unfold Rdiv; rewrite <- Rmult_opp_opp.
simpl (IZR 0).
apply Rmult_le_pos.
(replace 0 with (-ln 1); 
   [apply Ropp_le_contravar| rewrite ln_1; ring]).
apply Rlt_le; apply ln_increasing; intuition auto with real.
rewrite Ropp_inv_permute.
apply Rlt_le.
apply Rinv_0_lt_compat.
auto with real;fail.
auto with real.
assumption.
intros Hnepslt1.
exists 1%nat; simpl; rewrite Rmult_1_r.
assert (Heq: eps = 1).
apply Rle_antisym.
apply Rnot_lt_le; auto.
apply Rnot_lt_le; auto.
subst eps; intuition.

intros Hx0.
assert (Heq : x = 0).
apply Rle_antisym.
apply Rnot_lt_le; auto.
intuition.
subst x.
exists 1%nat; simpl; rewrite Rmult_1_r; assumption.
Qed.

Theorem GP_infinite :
  forall x,
    -1 < x < 1 ->
    x <> 0 ->
    infinit_sum (fun i => x^i)%R (1/(1-x)).
Proof.
intros x Hx1 Hx2.
assert (H1 : x <> 1).
apply Rlt_not_eq; dec_tac; fourier.
assert (H2 :  1 - x <> 0).
apply Rgt_not_eq; dec_tac; fourier.
unfold infinit_sum.
intros eps Heps_pos.
assert (H3 : 0 <= Rabs x < 1).
split; [apply Rabs_pos | apply Rabs_def1; dec_tac; fourier].
assert (Heps : 0 <eps*(1-x)).
dec_tac; apply Rmult_lt_0_compat; fourier.
elim (pow_lt_1 _ Heps _ H3).
intros N HN.
exists N%nat.
intros n Hn.
rewrite finite_GP; try assumption.
unfold R_dist.
replace ((1 - x ^ (n + 1)) / (1 - x) - 1 / (1 - x)) with 
  (- x ^ (n + 1)/ (1 - x)) ;[idtac| field; assumption].
rewrite Rabs_div ; try assumption.
rewrite Rabs_Ropp.
rewrite (Rabs_right (1-x)); try (dec_tac; fourier).
rewrite <- RPow_abs.
apply Rlt_div_r; [dec_tac; fourier| idtac].
apply Rle_lt_trans with (2:=HN).
apply pow_lt_1_decr; try omega.
split.
apply Rabs_pos_lt; assumption.
dec_tac; intuition.
Qed.

Theorem arctan_1_2_serie_bound :
  infinit_sum (fun i => (1/16)^i)%R (16/15).
Proof.
replace (16/15) with (1/ (1-1/16)).
apply GP_infinite.
split; fourier.
apply Rgt_not_eq; fourier.
field.
Qed.

Theorem GP_infinite_decomp :
  forall x k,
    -1 < x < 1 ->
    x <> 0 ->
    infinit_sum (fun i => x^(i+k))%R (x^k/(1-x)).
Proof.
intros x k Hx1 Hx2.
set (f :=(fun j => x^j)).
apply (infinit_sum_ext (fun i => f i * x^ k)).
intros n; unfold f.
symmetry; apply pow_add.
replace (x ^ k / (1 - x)) with (x ^ k * (1 / (1 - x))); 
  [idtac| unfold Rdiv; ring].
apply scal_infinit_sum.
unfold f.
apply GP_infinite; assumption.
apply pow_nonzero; assumption.
Qed.

Theorem plus_infinit_sum:
  forall (An Bn : nat -> R) s1 s2,
    infinit_sum An s1 ->
    infinit_sum Bn s2 ->
    infinit_sum (fun i : nat => An i + Bn i) (s1 + s2).
Proof.
intros An Bn s1 s2.
unfold infinit_sum.
intros H1 H2 eps Heps.
assert (Heps2 : eps/2 > 0) by fourier.
generalize (H1 (eps/2) Heps2).
generalize (H2 (eps/2) Heps2).
clear H1 H2.
intros [N1 HN1] [N2 HN2].

exists (N1+N2)%nat.
intros n Hn.
assert (H1 : (n >= N1)%nat) by omega.
assert (H2 : (n >= N2)%nat) by omega.
generalize (HN1 n H1).
generalize (HN2 n H2).
clear H1 H2 HN1 HN2.
intros H1 H2.
rewrite plus_sum.
apply Rle_lt_trans with ( R_dist (sum_f_R0 An n) s1 + R_dist (sum_f_R0 Bn n) s2).
apply R_dist_plus.
fourier.
Qed.

Theorem infinit_sum_opp : 
  forall An s, 
    infinit_sum An s -> 
    infinit_sum (fun i => - An i) (- s).
Proof.
intros An s Hinf.
apply infinit_sum_ext with (fun i => An i * -1);[ 
  intros n; ring|].
replace (-s) with (-1 * s) by ring.
apply scal_infinit_sum; [assumption| discrR].
Qed.

Theorem infinit_sum_pos :
  forall (An : nat -> R) s,
    (forall n : nat, 0 <= An n) -> 
    infinit_sum An s -> 
    0 <= s.
Proof.
intros An s H1 Hinf. 
apply (infinit_sum_bound (fun i:nat => 0) An); assumption || apply infinit_sum_0.
Qed.

Theorem infinit_sum_neg :
  forall (An : nat -> R) s,
    (forall n : nat, An n <= 0) -> 
    infinit_sum An s -> 
    s <= 0.
Proof.
intros An s H1 Hinf. 
assert ( 0 <= -s );[| fourier].
apply (infinit_sum_bound (fun i:nat => 0) (fun i => - An i)).
intros n; assert (H2:= H1 n); fourier.
apply infinit_sum_0.
apply infinit_sum_opp; assumption.
Qed.

Theorem finit_infinit_sum_bound : 
  forall An s,
    infinit_sum An s -> 
    (forall i, 0 <= An i) -> 
    forall k, sum_f_R0 An k <= s.
Proof.    
intros.    
assert (0 <= s - sum_f_R0 An k);[|fourier].
apply (infinit_sum_pos (fun i : nat => An (i + S k)%nat)).
auto.
apply infinit_sum_step_k; assumption.
Qed.

Theorem sum_f_split_odd : 
  forall An N,
    sum_f_R0 (fun i : nat => An (2 * i)%nat) N + 
    sum_f_R0 (fun i : nat => An (2 * i + 1)%nat) N  = 
    sum_f_R0 An (2*N + 1).
Proof.
intros An; induction N.
simpl; ring.
replace (2*S N+1)%nat with (S (S (2*N+1)))%nat by ring.
do 4 rewrite tech5.
rewrite <- IHN.
replace (2*S N+1)%nat with (S (S (2 * N+1)))%nat by ring.
replace (2 * S N)%nat with (S (2*N+1))%nat; ring.
Qed.

Theorem sum_f_split_even : 
  forall An N,
    (0 < N)%nat ->
    sum_f_R0 (fun i : nat => An (2 * i)%nat) N + 
    sum_f_R0 (fun i : nat => An (2 * i + 1)%nat) (N-1)%nat = 
    sum_f_R0 An (2*N)%nat.
Proof.
intros An N HN; induction N.
inversion HN.
cut (N=0 \/ 0 < N)%nat; [| omega].
intros H1; elim H1; clear H1; intros H1.
subst.
simpl; ring.
replace (S N -1)%nat with (S (N-1))%nat by omega.
replace (2*(S N))%nat with (S (S (2*N)))%nat by ring.
do 4 rewrite tech5.
rewrite <- IHN; try assumption.
replace (2*S (N-1)+1)%nat with (S (2 * N))%nat by omega.
replace (2 * S N)%nat with (S (S (2*N)))%nat; ring.
Qed.

Require Import Euclid.

Theorem infinit_sum_split : 
  forall An se so,
    infinit_sum (fun i => (An (2*i)%nat)) se -> 
    infinit_sum (fun i => (An (2*i+1)%nat)) so ->
    infinit_sum An (se+so)%R. 
Proof.	
unfold infinit_sum.
intros An se so He Ho eps Heps.
assert (Heps2 : (0 < eps/2)%R) by fourier.
generalize (He (eps/2) Heps2).
generalize (Ho (eps/2) Heps2).
clear He Ho.    
intros [No Ho] [Ne He].
exists (2*(No+Ne+1))%nat.
intros n Hn.
assert (U1 : (2>0)%nat) by omega.
elim (eucl_dev 2 U1 n)%nat.
intros q r H1 H2.
rewrite mult_comm in H2.
replace eps with (eps/2+eps/2) by field.
elim (zerop r); intros U2; subst.

rewrite plus_0_r.
assert (Hqo : ((q-1) >= No)%nat) by omega.
assert (Hqe : (q >= Ne)%nat) by omega.
generalize (He q Hqe); generalize (Ho (q-1)%nat Hqo); clear He Ho Hqe Hqo.
intros Ho He.
rewrite <- sum_f_split_even; try omega.
match goal with 
  |- R_dist (?a + ?c) (?b + ?d) < _ => 
    generalize (R_dist_plus a b c d); intros ; fourier
end.

assert (r=1)%nat; [omega| subst].
assert (Hqo : (q >= No)%nat) by omega.
assert (Hqe : (q >= Ne)%nat) by omega.
generalize (He q Hqe); generalize (Ho q Hqo); clear He Ho Hqe Hqo.
intros Ho He.
rewrite <- sum_f_split_odd; try omega.
match goal with 
  |- R_dist (?a + ?c) (?b + ?d) < _ => 
    generalize (R_dist_plus a b c d); intros ; fourier
end.
Qed.

Theorem infinit_sum_split2 : 
  forall An s1 s2 s3,
    infinit_sum (fun i => (An (2*i)%nat)) s1 -> 
    infinit_sum (fun i => (An (2*i+1)%nat)) s2 ->
    infinit_sum An s3 ->
    s1 + s2 = s3.
Proof.	
intros An s1 s2 s3 H1 H2.
apply uniqueness_sum.
apply infinit_sum_split; assumption.
Qed.

Definition finite_sum (Un : nat -> R) m n := 
  if (le_lt_dec n m) 
    then 0 
    else sum_f m (n -1)%nat Un.

Theorem infinit_sum_finite_sum :
  forall Un m n s, 
    (m <= n)%nat -> 
    infinit_sum (fun i => (Un (i + m)%nat)) s ->
    infinit_sum (fun i => (Un (i + n)%nat)) (s - finite_sum Un m n).
Proof.
intros Un m n s Hmn Hinf.
unfold finite_sum.
elim (le_lt_dec n m); intros G1.
assert (m = n) by omega; subst; clear G1.
replace (s - 0) with s by ring; assumption.
unfold sum_f.

assert (U1: (m < n)%nat) by omega; clear Hmn G1.
apply infinit_sum_ext with 
  (fun i => (fun j => (Un (j + m))) (i+ S (n - 1 - m)))%nat.  
  intros i.
  replace (i + S (n  - 1 - m) + m)%nat with (i + n)%nat by omega; reflexivity.
apply (infinit_sum_step_k (fun j => (Un (j + m)%nat))); assumption.
Qed.

Theorem finit_sum_split_3: 
  forall Un n,
    sum_f_R0 (fun n => (Un (2*n)%nat + Un (2*n+1)%nat)) n = 
    sum_f_R0 Un (2*n+1)%nat.
Proof.
intros Un n; induction n.
simpl; reflexivity.
replace (2 * S n + 1)%nat with (S (S (2*n+1)))%nat by ring.
simpl in *.
rewrite IHn.
replace (n + 0)%nat with n by ring.
replace (n + S n)%nat with (n+n+1)%nat by ring.
replace (n + n + 1 + 1)%nat with (S (n + n + 1))%nat; ring.
Qed.

Theorem infinit_sum_split3: 
  forall Un s,
    infinit_sum Un s ->
    infinit_sum (fun n => (Un (2*n)%nat + Un (2*n+1)%nat)) s.
Proof.
unfold infinit_sum.
intros Un s Hinf eps Heps.
destruct (Hinf eps Heps) as [N HN]; clear Hinf.
exists (2*N+1)%nat.
intros n HnN.
rewrite finit_sum_split_3. 
apply HN; omega.
Qed.

Theorem even_or_odd : 
  forall n, exists p, (n = 2*p \/ n = 2*p+1)%nat.
Proof.
intros n; induction n.
exists O.
left; ring.
destruct IHn as [p Hp].
destruct Hp as [Hp |Hp]; subst.
exists p; right; ring.
exists (S p); left; ring.
Qed.

Theorem infinit_sum_alt_pos : 
  forall Un s n,
    Un_decreasing Un ->
    infinit_sum (fun i => (tg_alt Un) (i + 2*n)%nat) s ->
    0 <= s.
Proof.
intros Un s n Hdecr Hinf.
assert (H1 := infinit_sum_split3 _ _ Hinf).
apply infinit_sum_pos with (2:= H1).
intros m.
unfold tg_alt.
replace (2 * m + 2 * n)%nat with (2* (m+n))%nat by ring.
replace (2 * m + 1 + 2 * n)%nat with (S (2* (m+n)))%nat by ring.
rewrite pow_1_even; rewrite pow_1_odd.
generalize (2*(m+n))%nat; intros p.
unfold Un_decreasing in Hdecr.
assert (H2:= Hdecr p); fourier.
Qed.

Theorem infinit_sum_alt_neg : 
  forall Un s n,
    Un_decreasing Un ->
    infinit_sum (fun i => (tg_alt Un) (i + 2*n+1)%nat) s ->
    s <= 0.
Proof.
intros Un s n Hdecr Hinf.
assert (H1 := infinit_sum_split3 _ _ Hinf).
apply infinit_sum_neg with (2:= H1).
intros m.
unfold tg_alt.
replace (2 * m + 2 * n+1)%nat with (S (2* (m+n)))%nat by ring.
replace (2 * m + 1 + 2 * n+1)%nat with ((2* S (m+n)))%nat by ring.
rewrite pow_1_even; rewrite pow_1_odd.
replace ((2* S (m+n)))%nat with (S (S (2* (m+n))))%nat by ring.
generalize (S (2*(m+n)))%nat; intros p.
unfold Un_decreasing in Hdecr.
assert (H2:= Hdecr p); fourier.
Qed.

Theorem  infinit_sum_alt_bound : 
  forall Un s n,
    Un_decreasing Un ->
    Un_cv Un 0 ->
    infinit_sum (fun i => (tg_alt Un) (i+n)%nat) s ->
    Rabs s <= Un n.
Proof.    
intros Un s n Hdecr Hcv Hinf.
assert (H1:= infinit_sum_step _ _ Hinf).
simpl in H1.
destruct (even_or_odd n) as [p Hp].
destruct Hp; subst.
  rewrite Rabs_right.
    assert (s - Un (2 * p)%nat <= 0); [| fourier].
    replace (Un (2 * p)%nat) with (tg_alt Un (2 * p)%nat).
      apply (infinit_sum_alt_neg Un) with p; [assumption|].
      apply infinit_sum_ext with (2:= H1).
      intros n; apply f_equal; ring.
      
      unfold tg_alt.
      rewrite pow_1_even; ring.
    apply Rle_ge.
    apply (infinit_sum_alt_pos Un s p); assumption.
    
  rewrite Rabs_left1.
    assert (0 <= s + Un (2 * p+1)%nat); [| fourier].
    replace (s+ Un (2 * p+1)%nat) with (s - tg_alt Un (2 * p+1)%nat).
      apply (infinit_sum_alt_pos Un) with (p+1)%nat; [assumption|].
      apply infinit_sum_ext with (2:= H1).
      intros n; apply f_equal; ring.
      
      unfold tg_alt.
      replace (2 * p + 1)%nat with (S (2*p))%nat by ring.
      rewrite pow_1_odd; ring.
      
    apply (infinit_sum_alt_neg Un s p); [assumption|].
    apply infinit_sum_ext with (2:= Hinf).
    intros n; apply f_equal; ring.
Qed.    


Theorem finite_sum_0 : 
  forall Un m,
    finite_sum Un m m = 0.
Proof.
intros Un m.
unfold finite_sum.
elim (le_lt_dec m m); intros G1.
auto.
absurd (m =m); omega.
Qed.

Theorem finite_sum_S : 
  forall Un m n, 
    (m <= n)%nat -> 
    finite_sum Un m (S n) = finite_sum Un m n + Un n. 
Proof.
intros Un m n Hmlen.
unfold finite_sum.
elim (le_lt_dec (S n) m); intros G1.
absurd (m =n); omega.
elim (le_lt_dec n m); intros G2.
assert (m=n) by omega; subst m.
replace (S n - 1)%nat with n by omega.
rewrite sum_f_single; ring.
replace (S n - 1)%nat with ((n-1)+1)%nat by omega.
rewrite sum_f_step_r; [| omega].
replace (n - 1 + 1)%nat with n by omega; ring.
Qed.

Theorem sum_f_R0_GP :
  forall x m n,
    (m < n)%nat ->
    x <> 1 ->
    sum_f_R0 (fun x0 : nat => x ^ (x0 + m)) (n - 1 - m) =
    x ^ m * (1 - x ^ (n - m)) / (1 - x).
Proof.
intros x m n Hmlen Hxneq1.
rewrite (sum_eq (fun x0 : nat => x ^ (x0 + m)) (fun i => x ^ i * x^m) (n - 1 - m)).
rewrite <- scal_sum.
rewrite (sum_eq (pow x) (fun i => x^i)).
rewrite finite_GP; [| assumption].
replace (n - 1 - m + 1)%nat with (n-m)%nat by omega.
unfold Rdiv; ring.
tauto.
intros; rewrite pow_add; reflexivity.
Qed.

Theorem sum_f_GP :
  forall x m n,
    (m < n)%nat ->
    x <> 1 ->
  sum_f m (n - 1) (pow x) = x ^ m * (1 - x ^ (n - m)) / (1 - x).
Proof.
intros x m n Hmltn Hx .
unfold sum_f.
apply sum_f_R0_GP; assumption.
Qed.

Theorem finite_sum_GP : 
  forall x m n,
    (m <= n)%nat ->
    x <> 1 -> 
    finite_sum (pow x) m n = x^m * (1- x^(n-m))/ (1-x).
Proof.
intros x m n Hmn Hxneq1.
unfold finite_sum.
elim (le_lt_dec n m); intros G1.
replace (n-m)%nat with 0%nat by omega.
unfold Rdiv; ring.
apply sum_f_GP; assumption.
Qed.



(*

Theorem infinit_sum_tg_bound_old : 
  forall Un s n,
    Un_decreasing Un ->
    Un_cv Un 0 ->
    infinit_sum (fun i => (tg_alt Un) (i+S n)%nat) s ->
    Rabs s <= Un (S n).
Proof.    
intros Un s n Hdecr Hconv Hinf.
destruct (alternated_series Un Hdecr Hconv) as [l H1].
change (Un_cv (fun N : nat => sum_f_R0 (tg_alt Un) N) l) with 
  (infinit_sum (tg_alt Un) l) in H1.

assert (R1 : s = (l-sum_f_R0 (tg_alt Un) n)).
  apply uniqueness_sum with (1:=Hinf).
  apply infinit_sum_step_k; assumption.

rewrite R1.
destruct (even_or_odd n) as [p Hp].
destruct Hp; subst.
 
  destruct (alternated_series_ineq _ _ p Hdecr Hconv H1) as [H2 H3].
  rewrite Rabs_left1; [| fourier] .
  assert (R2: Un (S (2 * p)) = 
    sum_f_R0 (tg_alt Un) (2 * p) - sum_f_R0 (tg_alt Un) (S (2 * p))).
    simpl sum_f_R0.
    replace (tg_alt Un (S (p + (p + 0)))) with (- Un (S (2 * p))); [ring|].
    unfold tg_alt.
    replace (p + (p + 0))%nat with (2*p)%nat by ring.
    rewrite pow_1_odd; ring.
  rewrite R2.
  replace (- (l - sum_f_R0 (tg_alt Un) (2 * p))) with 
    (sum_f_R0 (tg_alt Un) (2 * p) -l) by ring.
  fourier.

  destruct (alternated_series_ineq _ _ (p+1) Hdecr Hconv H1) as [H2 H3].
  rewrite Rabs_right.
    assert (R2: Un (S (2 * p+1)) = 
      sum_f_R0 (tg_alt Un) (S (2 * p+1)) - sum_f_R0 (tg_alt Un) (2 * p+1)).
      simpl sum_f_R0.
      replace (tg_alt Un (S (p + (p + 0)+1))) with (Un (S (2 * p+1))); [ring|].
      unfold tg_alt.
      replace (p + (p + 0)+1)%nat with (2*p+1)%nat by ring.
      replace (S (2 * p + 1))%nat with (2*(p+1))%nat by ring.
      rewrite pow_1_even; ring.
    rewrite R2.
    replace (S (2 * p + 1))%nat with (2 * (p + 1))%nat by ring.
    fourier.
  
  destruct (alternated_series_ineq _ _ p Hdecr Hconv H1) as [H4 _].
  replace (2 * p + 1)%nat with (S (2*p))%nat by ring.
  fourier.
Qed.


Theorem test_rabs : 
  forall x y, 
    x <> 0-> 
    -2 <= y / x <= 0 -> 
    Rabs (x + y) <= Rabs x.
Proof.
intros x y H1 H2.
apply Rdiv_le_r_1.
apply Rabs_pos_lt; assumption.
rewrite <- Rabs_div; [| assumption].
replace ((x + y) / x) with (1 + y / x) by (field; assumption).
rabs_elim.
generalize_all ( y / x ) z.
dec_tac; split; fourier.
Qed.

*)